# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Game tree structure for imperfect information games."""

import copy
from typing import Any, Dict, List, Text, Tuple

from open_spiel.python.examples.meta_cfr.sequential_games import cfr
from open_spiel.python.examples.meta_cfr.sequential_games import openspiel_api


class HistoryTreeNode:
  """Tree node to build game tree in cfr and do DFS traverse on game tree.

  Attributes:
    world_state: Current world state representation.
    reach_probs: Reach probability of tree node for each player. We consider
      reach probability for chance player, player 1 and player 2.
    action_probs: Probability of actions taken by each player. We consider
      actions taken by chance player, player 1 and player 2. Keys of this
      dictionary are tuples of (action_chance, action_player_1,
      action_player_2).
    children: A dictionary from a taken action from this node to the
      HistoryTreeNode of the child we derive in the game tree by taking an
      action.
  """

  def __init__(self, world_state: openspiel_api.WorldState):
    self.world_state = world_state
    self.reach_probs = [1.0, 1.0, 1.0]
    self.action_probs = {}
    self._value_p1 = 0
    self.children = {}

  def add_child(self, child_world_state: 'HistoryTreeNode',
                actions: Tuple[int, int, int]) -> None:
    """Adds the child world state to dictionary of children of this node."""
    self.children[actions] = child_world_state

  def get_child(self, actions: Tuple[int, int, int]) -> 'HistoryTreeNode':
    """Returns a child world state that can be derived from an action."""
    return self.children[actions]


class InfoState:
  """Information state class.

  Attributes:
    history_nodes: History of game as players play.
    player: Index of current player.
    infostate_string: String representation of current informantion state.
    world_state: Current game world state.
    children: Children nodes of information states. The keys are actions, and
      values are dictionary from information state string to information state
      node.
    counterfactual_reach_prob: Counterfactural values of reach probability for
      the current information state.
    player_reach_prob: Reach probability of information state for the acting
      player.
    counterfactual_action_values: Counterfactual values for each action in this
      information state. This is a dictionary from action to counterfactual
      value of this action in this information state.
    counterfactual_value: Counterfactual value of this information state.
    regret: Regret of each action for all player's actions in this information
      state.
    policy: Policy of player in this information state.
    average_policy: Average policy for all player's actions in this information
      state.
    average_policy_weight_sum: Sum of weighted average policy. This is used to
      normalize average policy and derive policy in this information state.
  """

  def __init__(self, world_state: openspiel_api.WorldState, player: int,
               infostate_string: Text):
    self.history_nodes = []
    self.player = player
    self.infostate_string = infostate_string
    self.world_state = world_state
    self._actions = world_state.get_actions()
    self.children = {a: {} for a in self._actions[player]}
    self.counterfactual_reach_prob = 0.
    self.player_reach_prob = 0.
    self.counterfactual_action_values = {}
    self.counterfactual_value = 0
    self.regret = {a: 0. for a in self._actions[player]}

    actions_count = len(self._actions[player])
    self.policy = {
        a: 1.0 / actions_count for a in world_state.get_actions()[player]
    }

    self.average_policy = {a: 0. for a in self._actions[player]}
    self.average_policy_weight_sum = 0.

  def add_history_node(self, history_node: HistoryTreeNode) -> None:
    """Updates history nodes with a given(last) history node."""
    self.history_nodes.append(history_node)

  def add_child_infostate(self, action: int,
                          infostate_child: Any) -> None:
    """Adds child infostate derived from taking an action to self.children."""
    self.children[action][infostate_child.infostate_string] = infostate_child

  def get_actions(self) -> List[int]:
    """Returns legal actions in current information state for current player."""
    return self.history_nodes[0].world_state.get_actions()[self.player]

  def is_terminal(self) -> bool:
    """Returns True if information state is terminal, False otherwise."""
    return self.history_nodes[0].world_state.is_terminal()


class GameTree:
  """Game tree class to build for CFR-based algorithms.

  Attributes:
    first_history_node: Root node of game tree.
    infostate_nodes: List of information state nodes for each player (including
      chance player).
    all_infostates_map: List of dictionaries (mapping from information state
      string representation to information state object) for each players
      (including chance player).
  """

  def __init__(self, first_history_node: HistoryTreeNode,
               infostate_nodes: List[InfoState],
               all_infostates_map: List[Dict[str, InfoState]]):
    self.first_history_node = first_history_node
    self.infostate_nodes = infostate_nodes
    self.all_infostates_map = all_infostates_map


def build_tree_dfs(
    world_state: openspiel_api.WorldState,
    all_infostates_map: List[Dict[str, InfoState]]
) -> Tuple[HistoryTreeNode, List[InfoState]]:
  """Builds the game tree by DFS traversal.

  Args:
    world_state: An openspiel game world state representation that will be the
      root of game tree.
    all_infostates_map: List of dictionaries (mapping from information state
      string representation to information state object) for each players
      (including chance player). This list will be empty when this function is
      called and it'll be population during DFS tree traversal.

  Returns:
    tree_node: Root of the game tree built in DFS traversal.
    infostate_nodes: List of information state (root) tree node for each player
    (including chance player).
  """
  tree_node = HistoryTreeNode(world_state)

  infostate_nodes = [
      InfoState(world_state, 1, world_state.get_infostate_string(1)),
      InfoState(world_state, 1, world_state.get_infostate_string(1)),
      InfoState(world_state, 2, world_state.get_infostate_string(2))
  ]
  for p in [cfr.Players.PLAYER_1, cfr.Players.PLAYER_2]:
    infostate_string = world_state.get_infostate_string(p)
    if infostate_string not in all_infostates_map[p]:
      all_infostates_map[p][infostate_string] = InfoState(
          world_state, p, infostate_string)

    infostate = all_infostates_map[p][infostate_string]
    infostate.add_history_node(tree_node)

    infostate_nodes[p] = infostate
  actions = world_state.get_actions()
  actions_chance, actions_p1, actions_p2 = actions

  for action_chance in actions_chance:
    for action_p1 in actions_p1:
      for action_p2 in actions_p2:
        child_state = copy.deepcopy(world_state)
        child_state.apply_actions((action_chance, action_p1, action_p2))
        child_tree_node, child_infostates = build_tree_dfs(
            child_state, all_infostates_map)

        tree_node.add_child(child_tree_node,
                            (action_chance, action_p1, action_p2))
        infostate_nodes[1].add_child_infostate(action_p1, child_infostates[1])
        infostate_nodes[2].add_child_infostate(action_p2, child_infostates[2])

  return tree_node, infostate_nodes


def build_game_tree(world_state: openspiel_api.WorldState) -> GameTree:
  """Builds game tree for CFR-based algorithms.

  Args:
    world_state: An openspiel game world state representation that will be the
      root of game tree.

  Returns:
    Calls GameTree function which returns the following:
    tree_node: Root of the game tree built in DFS traversal.
    infostate_nodes: List of information state (root) tree node for each player
    (including chance player).
  """
  all_infostates_map = [{}, {}, {}]
  first_history_node, infostate_nodes = build_tree_dfs(world_state,
                                                       all_infostates_map)
  return GameTree(first_history_node, infostate_nodes, all_infostates_map)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Neural Fictitious Self-Play (NFSP) agent implemented in TensorFlow.

See the paper https://arxiv.org/abs/1603.01121 for more details.
"""

import collections
import contextlib
import enum
import os
import random
from absl import logging
import numpy as np
import tensorflow.compat.v1 as tf

from open_spiel.python import rl_agent
from open_spiel.python import simple_nets
from open_spiel.python.algorithms import dqn

# Temporarily disable TF2 behavior until code is updated.
tf.disable_v2_behavior()

Transition = collections.namedtuple(
    "Transition", "info_state action_probs legal_actions_mask")

MODE = enum.Enum("mode", "best_response average_policy")


class NFSP(rl_agent.AbstractAgent):
  """NFSP Agent implementation in TensorFlow.

  See open_spiel/python/examples/kuhn_nfsp.py for an usage example.
  """

  def __init__(self,
               session,
               player_id,
               state_representation_size,
               num_actions,
               hidden_layers_sizes,
               reservoir_buffer_capacity,
               anticipatory_param,
               batch_size=128,
               rl_learning_rate=0.01,
               sl_learning_rate=0.01,
               min_buffer_size_to_learn=1000,
               learn_every=64,
               optimizer_str="sgd",
               **kwargs):
    """Initialize the `NFSP` agent."""
    self.player_id = player_id
    self._session = session
    self._num_actions = num_actions
    self._layer_sizes = hidden_layers_sizes
    self._batch_size = batch_size
    self._learn_every = learn_every
    self._anticipatory_param = anticipatory_param
    self._min_buffer_size_to_learn = min_buffer_size_to_learn

    self._reservoir_buffer = ReservoirBuffer(reservoir_buffer_capacity)
    self._prev_timestep = None
    self._prev_action = None

    # Step counter to keep track of learning.
    self._step_counter = 0

    # Inner RL agent
    kwargs.update({
        "batch_size": batch_size,
        "learning_rate": rl_learning_rate,
        "learn_every": learn_every,
        "min_buffer_size_to_learn": min_buffer_size_to_learn,
        "optimizer_str": optimizer_str,
    })
    self._rl_agent = dqn.DQN(session, player_id, state_representation_size,
                             num_actions, hidden_layers_sizes, **kwargs)

    # Keep track of the last training loss achieved in an update step.
    self._last_rl_loss_value = lambda: self._rl_agent.loss
    self._last_sl_loss_value = None

    # Placeholders.
    self._info_state_ph = tf.placeholder(
        shape=[None, state_representation_size],
        dtype=tf.float32,
        name="info_state_ph")

    self._action_probs_ph = tf.placeholder(
        shape=[None, num_actions], dtype=tf.float32, name="action_probs_ph")

    self._legal_actions_mask_ph = tf.placeholder(
        shape=[None, num_actions],
        dtype=tf.float32,
        name="legal_actions_mask_ph")

    # Average policy network.
    self._avg_network = simple_nets.MLP(state_representation_size,
                                        self._layer_sizes, num_actions)
    self._avg_policy = self._avg_network(self._info_state_ph)
    self._avg_policy_probs = tf.nn.softmax(self._avg_policy)

    self._savers = [
        ("q_network", tf.train.Saver(self._rl_agent._q_network.variables)),
        ("avg_network", tf.train.Saver(self._avg_network.variables))
    ]

    # Loss
    self._loss = tf.reduce_mean(
        tf.nn.softmax_cross_entropy_with_logits_v2(
            labels=tf.stop_gradient(self._action_probs_ph),
            logits=self._avg_policy))

    if optimizer_str == "adam":
      optimizer = tf.train.AdamOptimizer(learning_rate=sl_learning_rate)
    elif optimizer_str == "sgd":
      optimizer = tf.train.GradientDescentOptimizer(
          learning_rate=sl_learning_rate)
    else:
      raise ValueError("Not implemented. Choose from ['adam', 'sgd'].")

    self._learn_step = optimizer.minimize(self._loss)
    self._sample_episode_policy()

  @contextlib.contextmanager
  def temp_mode_as(self, mode):
    """Context manager to temporarily overwrite the mode."""
    previous_mode = self._mode
    self._mode = mode
    yield
    self._mode = previous_mode

  def get_step_counter(self):
    return self._step_counter

  def _sample_episode_policy(self):
    if np.random.rand() < self._anticipatory_param:
      self._mode = MODE.best_response
    else:
      self._mode = MODE.average_policy

  def _act(self, info_state, legal_actions):
    info_state = np.reshape(info_state, [1, -1])
    action_values, action_probs = self._session.run(
        [self._avg_policy, self._avg_policy_probs],
        feed_dict={self._info_state_ph: info_state})

    self._last_action_values = action_values[0]
    # Remove illegal actions, normalize probs
    probs = np.zeros(self._num_actions)
    probs[legal_actions] = action_probs[0][legal_actions]
    probs /= sum(probs)
    action = np.random.choice(len(probs), p=probs)
    return action, probs

  @property
  def mode(self):
    return self._mode

  @property
  def loss(self):
    return (self._last_sl_loss_value, self._last_rl_loss_value())

  def step(self, time_step, is_evaluation=False):
    """Returns the action to be taken and updates the Q-networks if needed.

    Args:
      time_step: an instance of rl_environment.TimeStep.
      is_evaluation: bool, whether this is a training or evaluation call.

    Returns:
      A `rl_agent.StepOutput` containing the action probs and chosen action.
    """
    if self._mode == MODE.best_response:
      agent_output = self._rl_agent.step(time_step, is_evaluation)
      if not is_evaluation and not time_step.last():
        self._add_transition(time_step, agent_output)

    elif self._mode == MODE.average_policy:
      # Act step: don't act at terminal info states.
      if not time_step.last():
        info_state = time_step.observations["info_state"][self.player_id]
        legal_actions = time_step.observations["legal_actions"][self.player_id]
        action, probs = self._act(info_state, legal_actions)
        agent_output = rl_agent.StepOutput(action=action, probs=probs)

      if self._prev_timestep and not is_evaluation:
        self._rl_agent.add_transition(self._prev_timestep, self._prev_action,
                                      time_step)
    else:
      raise ValueError("Invalid mode ({})".format(self._mode))

    if not is_evaluation:
      self._step_counter += 1

      if self._step_counter % self._learn_every == 0:
        self._last_sl_loss_value = self._learn()
        # If learn step not triggered by rl policy, learn.
        if self._mode == MODE.average_policy:
          self._rl_agent.learn()

      # Prepare for the next episode.
      if time_step.last():
        self._sample_episode_policy()
        self._prev_timestep = None
        self._prev_action = None
        return
      else:
        self._prev_timestep = time_step
        self._prev_action = agent_output.action

    return agent_output

  def _add_transition(self, time_step, agent_output):
    """Adds the new transition using `time_step` to the reservoir buffer.

    Transitions are in the form (time_step, agent_output.probs, legal_mask).

    Args:
      time_step: an instance of rl_environment.TimeStep.
      agent_output: an instance of rl_agent.StepOutput.
    """
    legal_actions = time_step.observations["legal_actions"][self.player_id]
    legal_actions_mask = np.zeros(self._num_actions)
    legal_actions_mask[legal_actions] = 1.0
    transition = Transition(
        info_state=(time_step.observations["info_state"][self.player_id][:]),
        action_probs=agent_output.probs,
        legal_actions_mask=legal_actions_mask)
    self._reservoir_buffer.add(transition)

  def _learn(self):
    """Compute the loss on sampled transitions and perform a avg-network update.

    If there are not enough elements in the buffer, no loss is computed and
    `None` is returned instead.

    Returns:
      The average loss obtained on this batch of transitions or `None`.
    """
    if (len(self._reservoir_buffer) < self._batch_size or
        len(self._reservoir_buffer) < self._min_buffer_size_to_learn):
      return None

    transitions = self._reservoir_buffer.sample(self._batch_size)
    info_states = [t.info_state for t in transitions]
    action_probs = [t.action_probs for t in transitions]
    legal_actions_mask = [t.legal_actions_mask for t in transitions]

    loss, _ = self._session.run(
        [self._loss, self._learn_step],
        feed_dict={
            self._info_state_ph: info_states,
            self._action_probs_ph: action_probs,
            self._legal_actions_mask_ph: legal_actions_mask,
        })
    return loss

  def _full_checkpoint_name(self, checkpoint_dir, name):
    checkpoint_filename = "_".join([name, "pid" + str(self.player_id)])
    return os.path.join(checkpoint_dir, checkpoint_filename)

  def _latest_checkpoint_filename(self, name):
    checkpoint_filename = "_".join([name, "pid" + str(self.player_id)])
    return checkpoint_filename + "_latest"

  def save(self, checkpoint_dir):
    """Saves the average policy network and the inner RL agent's q-network.

    Note that this does not save the experience replay buffers and should
    only be used to restore the agent's policy, not resume training.

    Args:
      checkpoint_dir: directory where checkpoints will be saved.
    """
    for name, saver in self._savers:
      path = saver.save(
          self._session,
          self._full_checkpoint_name(checkpoint_dir, name),
          latest_filename=self._latest_checkpoint_filename(name))
      logging.info("Saved to path: %s", path)

  def has_checkpoint(self, checkpoint_dir):
    for name, _ in self._savers:
      if tf.train.latest_checkpoint(
          self._full_checkpoint_name(checkpoint_dir, name),
          os.path.join(checkpoint_dir,
                       self._latest_checkpoint_filename(name))) is None:
        return False
    return True

  def restore(self, checkpoint_dir):
    """Restores the average policy network and the inner RL agent's q-network.

    Note that this does not restore the experience replay buffers and should
    only be used to restore the agent's policy, not resume training.

    Args:
      checkpoint_dir: directory from which checkpoints will be restored.
    """
    for name, saver in self._savers:
      full_checkpoint_dir = self._full_checkpoint_name(checkpoint_dir, name)
      logging.info("Restoring checkpoint: %s", full_checkpoint_dir)
      saver.restore(self._session, full_checkpoint_dir)


class ReservoirBuffer(object):
  """Allows uniform sampling over a stream of data.

  This class supports the storage of arbitrary elements, such as observation
  tensors, integer actions, etc.

  See https://en.wikipedia.org/wiki/Reservoir_sampling for more details.
  """

  def __init__(self, reservoir_buffer_capacity):
    self._reservoir_buffer_capacity = reservoir_buffer_capacity
    self._data = []
    self._add_calls = 0

  def add(self, element):
    """Potentially adds `element` to the reservoir buffer.

    Args:
      element: data to be added to the reservoir buffer.
    """
    if len(self._data) < self._reservoir_buffer_capacity:
      self._data.append(element)
    else:
      idx = np.random.randint(0, self._add_calls + 1)
      if idx < self._reservoir_buffer_capacity:
        self._data[idx] = element
    self._add_calls += 1

  def sample(self, num_samples):
    """Returns `num_samples` uniformly sampled from the buffer.

    Args:
      num_samples: `int`, number of samples to draw.

    Returns:
      An iterable over `num_samples` random elements of the buffer.

    Raises:
      ValueError: If there are less than `num_samples` elements in the buffer
    """
    if len(self._data) < num_samples:
      raise ValueError("{} elements could not be sampled from size {}".format(
          num_samples, len(self._data)))
    return random.sample(self._data, num_samples)

  def clear(self):
    self._data = []
    self._add_calls = 0

  def __len__(self):
    return len(self._data)

  def __iter__(self):
    return iter(self._data)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.adidas_utils.helpers.simplex."""

from absl.testing import absltest
from absl.testing import parameterized

import numpy as np

from open_spiel.python.algorithms.adidas_utils.helpers import simplex


class SimplexTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('inside', np.array([.25, .75]), np.array([.25, .75])),
      ('outside_1', np.ones(2), 0.5 * np.ones(2)),
      ('outside_2', np.array([2., 0.]), np.array([1., 0.])),
      ('outside_3', np.array([.25, .25]), np.array([.5, .5])),
      )
  def test_euclidean_projection(self, vector, expected_projection):
    projection = simplex.euclidean_projection_onto_simplex(vector, subset=False)
    self.assertListEqual(list(projection), list(expected_projection),
                         msg='projection not accurate')

  @parameterized.named_parameters(
      ('orth', np.array([.75, .75]), np.array([.0, .0])),
      ('oblique', np.array([1., .5]), np.array([.25, -.25])),
      ('tangent', np.array([.25, .25, -.5]), np.array([.25, .25, -.5])),
      )
  def test_tangent_projection(self, vector, expected_projection):
    projection = simplex.project_grad(vector)
    self.assertListEqual(list(projection), list(expected_projection),
                         msg='projection not accurate')

  @parameterized.named_parameters(
      ('orth_1', np.array([0.5, 0.5]), np.array([.75, .75]), 0.),
      ('orth_2', np.array([1., 0.]), np.array([.75, .75]), 0.),
      ('tangent_1', np.array([1., 0.]), np.array([-.5, .5]), 0.),
      ('tangent_2', np.array([1., 0.]), np.array([1., -1.]), np.sqrt(2)),
      )
  def test_grad_norm(self, dist, grad, expected_norm):
    norm = simplex.grad_norm(dist, grad)
    self.assertAlmostEqual(norm, expected_norm, msg='norm not accurate')


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.egt.utils."""

import itertools
from absl.testing import absltest

from absl.testing import parameterized
import numpy as np

from open_spiel.python.egt import utils
import pyspiel


def _generate_prob_profiles(num_items, num_slots):
  """Another implementation of `distribution` for test purposes.

  This function is the original implementation from Karl. jblespiau@ find it
  useful to add it here as: 1) an additional test of our function 2) a check
  that the initial code is correct too.

  Args:
    num_items: The number of items to distribute.
    num_slots: The number of slots.

  Returns:
    A numpy array of shape [num_distributions, num_slots].
  """
  if num_slots == 1:
    return np.array([num_items])

  num_rows = utils.n_choose_k(num_items + num_slots - 1, num_items)
  distributions = np.empty([num_rows, num_slots])

  ind = 0
  for i in range(0, num_items + 1):
    n_tmp = num_items - i
    k_tmp = num_slots - 1
    distributions_tmp = _generate_prob_profiles(n_tmp, k_tmp)
    distributions[ind:ind +
                  np.shape(distributions_tmp)[0], :] = np.column_stack(
                      (np.array((np.ones(np.shape(distributions_tmp)[0]) * i)),
                       distributions_tmp))
    ind = ind + np.shape(distributions_tmp)[0]

  return distributions


class UtilsTest(parameterized.TestCase):

  @parameterized.parameters(
      (5, 3, False),
      (2, 2, True),
  )
  def test_distribution(self, num_items, num_slots, normalize):
    distribution = list(utils.distribute(num_items, num_slots, normalize))
    # Correct length.
    # See https://en.wikipedia.org/wiki/Stars_and_bars_%28combinatorics%29.
    self.assertLen(distribution,
                   utils.n_choose_k(num_items + num_slots - 1, num_items))
    # No duplicates.
    self.assertLen(distribution, len(set(distribution)))
    sum_distribution = num_items if not normalize else 1
    for d in distribution:
      self.assertTrue(sum_distribution, sum(d))
      self.assertTrue((np.asarray(d) >= 0).all())

  @parameterized.parameters(
      (5, 3),
      (2, 2),
      (3, 3),
      (10, 5),
  )
  def test_distribution_equivalent_implementation(self, num_items, num_slots):
    dist_list = list(utils.distribute(num_items, num_slots, normalize=False))
    distribution = np.vstack(dist_list)

    other_implementation = _generate_prob_profiles(num_items, num_slots)
    np.testing.assert_array_equal(
        utils.sort_rows_lexicographically(distribution),
        utils.sort_rows_lexicographically(other_implementation))

  def test_sort_rows_lexicographically(self):
    array = np.asarray([
        [1, 1, 0],
        [1, 2, 0],
        [3, 1, 0],
        [0, 0, 4],
    ])
    expected = np.asarray([
        [0, 0, 4],
        [1, 1, 0],
        [1, 2, 0],
        [3, 1, 0],
    ])

    np.testing.assert_equal(expected, utils.sort_rows_lexicographically(array))

  def test_id_profile_mapping(self):
    """Tests forward and backward mapping of pure strategy profiles to IDs."""

    num_strats_per_population = np.asarray([4, 4, 4, 9])
    num_pure_profiles = np.prod(num_strats_per_population)

    strat_ranges = [
        range(num_strats) for num_strats in num_strats_per_population
    ]

    id_list = []
    for strat_profile in itertools.product(strat_ranges[0], strat_ranges[1],
                                           strat_ranges[2], strat_ranges[3]):
      profile_id = utils.get_id_from_strat_profile(num_strats_per_population,
                                                   strat_profile)
      id_list.append(profile_id)

      # Tests backward mapping (ID-to-profile lookup)
      strat_profile_from_id = utils.get_strat_profile_from_id(
          num_strats_per_population, profile_id)
      np.testing.assert_array_equal(strat_profile, strat_profile_from_id)

    # Tests forward mapping (profile-to-ID lookup)
    np.testing.assert_array_equal(id_list, range(num_pure_profiles))

  def test_get_valid_next_profiles(self):
    """Tests next-profile generator."""

    num_strats_per_population = np.asarray([4, 5, 9, 7])
    cur_profile = np.asarray([1, 1, 2, 1])
    next_profiles = utils.get_valid_next_profiles(num_strats_per_population,
                                                  cur_profile)

    num_next_profiles = 0
    for _, _ in next_profiles:
      num_next_profiles += 1

    expected = (num_strats_per_population - 1).sum()
    np.testing.assert_equal(expected, num_next_profiles)

  def test_constant_sum_checker(self):
    """Tests if verification of constant-sum game is correct."""

    game = pyspiel.load_matrix_game("matrix_rps")
    payoff_tables = utils.game_payoffs_array(game)
    payoffs_are_hpt_format = utils.check_payoffs_are_hpt(payoff_tables)
    game_is_constant_sum, payoff_sum = utils.check_is_constant_sum(
        payoff_tables[0], payoffs_are_hpt_format)
    self.assertTrue(game_is_constant_sum)
    self.assertEqual(payoff_sum, 0.)

  def test_game_payoffs_array_rps(self):
    """Test `game_payoffs_array` for rock-paper-scissors."""
    game = pyspiel.load_matrix_game("matrix_rps")
    payoff_matrix = np.empty(shape=(2, 3, 3))
    payoff_row = np.array([[0., -1., 1.], [1., 0., -1.], [-1., 1., 0.]])
    payoff_matrix[0] = payoff_row
    payoff_matrix[1] = -1. * payoff_row
    np.testing.assert_allclose(utils.game_payoffs_array(game), payoff_matrix)

  def test_game_payoffs_array_pd(self):
    """Test `game_payoffs_array` for prisoners' dilemma."""
    game = pyspiel.load_matrix_game("matrix_pd")
    payoff_matrix = np.empty(shape=(2, 2, 2))
    payoff_row = np.array([[5., 0.], [10., 1.]])
    payoff_matrix[0] = payoff_row
    payoff_matrix[1] = payoff_row.T
    np.testing.assert_allclose(utils.game_payoffs_array(game), payoff_matrix)

  @parameterized.parameters(
      (100, 2, 0.),
      (100, 3, 0.),
      (100, 4, 0.),
      (100, 2, 0.05),
  )
  def test_sample_from_simplex(self, n, dim, vmin):
    """Test `sample_from_simplex`."""
    x = utils.sample_from_simplex(n, dim=dim, vmin=vmin)
    np.testing.assert_allclose(np.sum(x, axis=1), np.ones(n))
    self.assertTrue(np.all(x <= 1. - vmin))
    self.assertTrue(np.all(x >= vmin))


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for iterated_prisoners_dilemma.py."""

from absl.testing import absltest

from open_spiel.python.games import iterated_prisoners_dilemma  # pylint: disable=unused-import
import pyspiel


class IteratedPrisonersDilemmaTest(absltest.TestCase):

  def test_default_param(self):
    """Check the game can be converted to a turn-based game."""
    game = pyspiel.load_game("python_iterated_prisoners_dilemma")
    self.assertEqual(game._termination_probability, 0.125)

  def test_non_default_param_from_string(self):
    """Check params can be given through the string loading."""
    game = pyspiel.load_game(
        "python_iterated_prisoners_dilemma(termination_probability=0.5)")
    self.assertEqual(game._termination_probability, 0.5)

  def test_non_default_param_from_dict(self):
    """Check params can be given through a dictionary."""
    game = pyspiel.load_game("python_iterated_prisoners_dilemma",
                             {"termination_probability": 0.75})
    self.assertEqual(game._termination_probability, 0.75)

  def test_game_as_turn_based(self):
    """Check the game can be converted to a turn-based game."""
    game = pyspiel.load_game("python_iterated_prisoners_dilemma")
    turn_based = pyspiel.convert_to_turn_based(game)
    pyspiel.random_sim_test(
        turn_based, num_sims=10, serialize=False, verbose=True)

  def test_game_as_turn_based_via_string(self):
    """Check the game can be created as a turn-based game from a string."""
    game = pyspiel.load_game(
        "turn_based_simultaneous_game(game=python_iterated_prisoners_dilemma())"
    )
    pyspiel.random_sim_test(game, num_sims=10, serialize=False, verbose=True)

  def test_game_from_cc(self):
    """Runs our standard game tests, checking API consistency."""
    game = pyspiel.load_game("python_iterated_prisoners_dilemma")
    pyspiel.random_sim_test(game, num_sims=10, serialize=False, verbose=True)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests the C++ nfg_game methods exposed to Python."""

from absl.testing import absltest

import pyspiel


class NFGGameTest(absltest.TestCase):

  def test_pd(self):
    pd_nfg_string = ("""NFG 1 R "OpenSpiel export of matrix_pd()"
{ "Player 0" "Player 1" } { 2 2 }

5 5
10 0
0 10
1 1
""")
    game = pyspiel.load_nfg_game(pd_nfg_string)
    # First (row) player utilities (player, row, col)
    self.assertEqual(game.player_utility(0, 0, 0), 5)
    self.assertEqual(game.player_utility(0, 1, 0), 10)
    self.assertEqual(game.player_utility(0, 0, 1), 0)
    self.assertEqual(game.player_utility(0, 1, 1), 1)
    # Now, second (column) player
    self.assertEqual(game.player_utility(1, 0, 0), 5)
    self.assertEqual(game.player_utility(1, 1, 0), 0)
    self.assertEqual(game.player_utility(1, 0, 1), 10)
    self.assertEqual(game.player_utility(1, 1, 1), 1)

  def test_native_export_import(self):
    """Check that we can import games that we've exported.

    We do not do any additional checking here, as these methods are already
    being extensively tested in nfg_test.cc. The purpose of this test is only
    to check that the python wrapping works.
    """
    game_strings = [
        "matrix_rps", "matrix_shapleys_game", "matrix_pd", "matrix_sh",
        "blotto(players=2,coins=5,fields=3)",
        "blotto(players=3,coins=5,fields=3)"
    ]
    for game_string in game_strings:
      game = pyspiel.load_game(game_string)
      nfg_text = pyspiel.game_to_nfg_string(game)
      nfg_game = pyspiel.load_nfg_game(nfg_text)
      self.assertIsNotNone(nfg_game)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements an Ephemeral Value Adjustment Agent.

See https://arxiv.org/abs/1810.08163.

The algorithm queries trajectories from a replay buffer based on similarities
to embedding representations and uses a parametric model to compute values for
counterfactual state-action pairs when integrating across those trajectories.
Finally, a weighted average between the parametric (DQN in this case) and the
non-parametric model is used to compute the policy.
"""

import collections
import copy
import numpy as np
import tensorflow.compat.v1 as tf

from open_spiel.python import rl_agent
from open_spiel.python import simple_nets
from open_spiel.python.algorithms import dqn

# Temporarily disable TF2 behavior until we update the code.
tf.disable_v2_behavior()

MEM_KEY_NAME = "embedding"

ValueBufferElement = collections.namedtuple("ValueElement", "embedding value")

ReplayBufferElement = collections.namedtuple(
    "ReplayElement", "embedding info_state action reward next_info_state "
    "is_final_step legal_actions_mask")


# TODO(author3) Refactor into data structures lib.
class QueryableFixedSizeRingBuffer(dqn.ReplayBuffer):
  """ReplayBuffer of fixed size with a FIFO replacement policy.

  Stored transitions can be sampled uniformly.  This extends the DQN replay
  buffer by allowing the contents to be fetched by L2 proximity to a query
  value.

  The underlying datastructure is a ring buffer, allowing 0(1) adding and
  sampling.
  """

  def knn(self, key, key_name, k, trajectory_len=1):
    """Computes top-k neighbours based on L2 distance.

    Args:
      key: (np.array) key value to query memory.
      key_name:  (str) attribute name of key in memory elements.
      k: (int) number of neighbours to fetch.
      trajectory_len: (int) length of trajectory to fetch from replay buffer.

    Returns:
      List of tuples (L2 negative distance, BufferElement) sorted in increasing
      order by the negative L2 distqances  from the key.
    """
    distances = [(np.linalg.norm(getattr(sample, key_name) - key, 2,
                                 axis=0), sample) for sample in self._data]
    return sorted(distances, key=lambda v: -v[0])[:k]


class EVAAgent(object):
  """Implements a solver for Ephemeral VAlue Adjustment.

  See https://arxiv.org/abs/1810.08163.

  Define all networks and sampling buffers/memories.  Derive losses & learning
  steps. Initialize the game state and algorithmic variables.
  """

  def __init__(self,
               session,
               game,
               player_id,
               state_size,
               num_actions,
               embedding_network_layers=(128,),
               embedding_size=16,
               dqn_hidden_layers=(128, 128),
               batch_size=16,
               trajectory_len=10,
               num_neighbours=5,
               learning_rate=1e-4,
               mixing_parameter=0.9,
               memory_capacity=int(1e6),
               discount_factor=1.0,
               update_target_network_every=1000,
               epsilon_start=1.0,
               epsilon_end=0.1,
               epsilon_decay_duration=int(1e4),
               embedding_as_parametric_input=False):
    """Initialize the Ephemeral VAlue Adjustment algorithm.

    Args:
      session: (tf.Session) TensorFlow session.
      game: (rl_environment.Environment) Open Spiel game.
      player_id: (int) Player id for this player.
      state_size: (int) Size of info state vector.
      num_actions: (int) number of actions.
      embedding_network_layers: (list[int]) Layer sizes of strategy net MLP.
      embedding_size: (int) Size of memory embeddings.
      dqn_hidden_layers: (list(int)) MLP layer sizes of DQN network.
      batch_size: (int) Size of batches for DQN learning steps.
      trajectory_len: (int) Length of trajectories from replay buffer.
      num_neighbours: (int) Number of neighbours to fetch from replay buffer.
      learning_rate: (float) Learning rate.
      mixing_parameter: (float) Value mixing parameter between 0 and 1.
      memory_capacity: Number af samples that can be stored in memory.
      discount_factor: (float) Discount factor for Q-Learning.
      update_target_network_every: How often to update DQN target network.
      epsilon_start: (float) Starting epsilon-greedy value.
      epsilon_end: (float) Final epsilon-greedy value.
      epsilon_decay_duration: (float) Number of steps over which epsilon decays.
      embedding_as_parametric_input: (bool) Whether we use embeddings as input
        to the parametric model.
    """
    assert (mixing_parameter >= 0 and mixing_parameter <= 1)
    self._game = game
    self._session = session
    self.player_id = player_id
    self._env = game
    self._num_actions = num_actions
    self._info_state_size = state_size
    self._embedding_size = embedding_size
    self._lambda = mixing_parameter
    self._trajectory_len = trajectory_len
    self._num_neighbours = num_neighbours
    self._discount = discount_factor
    self._epsilon_start = epsilon_start
    self._epsilon_end = epsilon_end
    self._epsilon_decay_duration = epsilon_decay_duration
    self._last_time_step = None
    self._last_action = None
    self._embedding_as_parametric_input = embedding_as_parametric_input

    # Create required TensorFlow placeholders to perform the Q-network updates.
    self._info_state_ph = tf.placeholder(
        shape=[None, self._info_state_size],
        dtype=tf.float32,
        name="info_state_ph")
    self._embedding_network = simple_nets.MLP(self._info_state_size,
                                              list(embedding_network_layers),
                                              embedding_size)
    self._embedding = self._embedding_network(self._info_state_ph)

    # The DQN agent requires this be an integer.
    if not isinstance(memory_capacity, int):
      raise ValueError("Memory capacity not an integer.")

    # Initialize the parametric & non-parametric Q-networks.
    self._agent = dqn.DQN(
        session,
        player_id,
        state_representation_size=self._info_state_size,
        num_actions=self._num_actions,
        hidden_layers_sizes=list(dqn_hidden_layers),
        replay_buffer_capacity=memory_capacity,
        replay_buffer_class=QueryableFixedSizeRingBuffer,
        batch_size=batch_size,
        learning_rate=learning_rate,
        update_target_network_every=update_target_network_every,
        learn_every=batch_size,
        discount_factor=1.0,
        epsilon_start=1.0,
        epsilon_end=0.1,
        epsilon_decay_duration=int(1e6))
    # Initialize Value Buffers - Fetch Replay buffers from agents.
    self._value_buffer = QueryableFixedSizeRingBuffer(memory_capacity)
    self._replay_buffer = self._agent.replay_buffer

    # Initialize non-parametric & EVA Q-values.
    self._v_np = collections.defaultdict(float)
    self._q_np = collections.defaultdict(lambda: [0] * self._num_actions)
    self._q_eva = collections.defaultdict(lambda: [0] * self._num_actions)

  @property
  def env(self):
    return self._env

  @property
  def loss(self):
    return self._agent.loss

  def _add_transition_value(self, infostate_embedding, value):
    """Adds the embedding and value to the ValueBuffer.

    Args:
      infostate_embedding: (np.array) embeddig vector.
      value: (float) Value associated with state embeding.
    """
    transition = ValueBufferElement(embedding=infostate_embedding, value=value)
    self._value_buffer.add(transition)

  def _add_transition_replay(self, infostate_embedding, time_step):
    """Adds the new transition using `time_step` to the replay buffer.

    Adds the transition from `self._prev_timestep` to `time_step` by
    `self._prev_action`.

    Args:
      infostate_embedding: embeddig vector.
      time_step: an instance of rl_environment.TimeStep.
    """
    prev_timestep = self._last_time_step
    assert prev_timestep is not None
    legal_actions = (
        prev_timestep.observations["legal_actions"][self.player_id])
    legal_actions_mask = np.zeros(self._num_actions)
    legal_actions_mask[legal_actions] = 1.0
    reward = time_step.rewards[self.player_id] if time_step.rewards else 0.0
    transition = ReplayBufferElement(
        embedding=infostate_embedding,
        info_state=(prev_timestep.observations["info_state"][self.player_id]),
        action=self._last_action,
        reward=reward,
        next_info_state=time_step.observations["info_state"][self.player_id],
        is_final_step=float(time_step.last()),
        legal_actions_mask=legal_actions_mask)
    self._replay_buffer.add(transition)

  def step(self, time_step, is_evaluation=False):
    """Returns the action to be taken and updates the value functions.

    Args:
      time_step: an instance of rl_environment.TimeStep.
      is_evaluation: bool, whether this is a training or evaluation call.

    Returns:
      A `rl_agent.StepOutput` containing the action probs and chosen action.
    """
    # Act step: don't act at terminal info states.
    if not time_step.last():
      info_state = time_step.observations["info_state"][self.player_id]
      legal_actions = time_step.observations["legal_actions"][self.player_id]
      epsilon = self._get_epsilon(self._agent.step_counter, is_evaluation)

      # Sample an action from EVA via epsilon greedy policy.
      action, probs = self._epsilon_greedy(self._q_eva[tuple(info_state)],
                                           legal_actions, epsilon)

    # Update Step: Only with transitions and not when evaluating.
    if (not is_evaluation and self._last_time_step is not None):
      info_state = self._last_time_step.observations["info_state"][
          self.player_id]
      legal_actions = self._last_time_step.observations["legal_actions"][
          self.player_id]
      epsilon = self._get_epsilon(self._agent.step_counter, is_evaluation)

      # Get embedding.
      infostate_embedding = self._session.run(
          self._embedding,
          feed_dict={self._info_state_ph: np.expand_dims(info_state,
                                                         axis=0)})[0]

      neighbours_value = self._value_buffer.knn(infostate_embedding,
                                                MEM_KEY_NAME,
                                                self._num_neighbours, 1)
      # collect trace values of knn from L (value buffer) .. Q_np(s_k)
      neighbours_replay = self._replay_buffer.knn(infostate_embedding,
                                                  MEM_KEY_NAME,
                                                  self._num_neighbours,
                                                  self._trajectory_len)

      # Take a step with the parametric model and get q-values. Use embedding as
      # input to the parametric meodel.
      # TODO(author6) Recompute embeddings for buffers on learning steps.
      if self._embedding_as_parametric_input:
        last_time_step_copy = copy.deepcopy(self._last_time_step)
        last_time_step_copy.observations["info_state"][
            self.player_id] = infostate_embedding
        self._agent.step(last_time_step_copy, add_transition_record=False)
      else:
        self._agent.step(self._last_time_step, add_transition_record=False)
      q_values = self._session.run(
          self._agent.q_values,
          feed_dict={
              self._agent.info_state_ph: np.expand_dims(info_state, axis=0)
          })[0]
      # Update EVA: Q_eva = lambda q_theta(s_t) + (1-lambda) sum(Q_np(s_k, .))/K
      for a in legal_actions:
        q_theta = q_values[a]
        self._q_eva[tuple(info_state)][a] = (
            self._lambda * q_theta + (1 - self._lambda) *
            sum([elem[1].value
                 for elem in neighbours_value]) / self._num_neighbours)

      # Append (e,s,a,r,s') to Replay Buffer
      self._add_transition_replay(infostate_embedding, time_step)

      # update Q_np with Traces using TCP
      self._trajectory_centric_planning(neighbours_replay)

      # Append Q_np(s, a) to Value Buffer
      self._add_transition_value(
          infostate_embedding, self._q_np[tuple(info_state)][self._last_action])

    # Prepare for the next episode.
    if time_step.last():
      self._last_time_step = None
      self._last_action = None
      return

    self._last_time_step = time_step
    self._last_action = action
    return rl_agent.StepOutput(action=action, probs=probs)

  def _trajectory_centric_planning(self, trajectories):
    """Performs trajectory centric planning.

    Uses trajectories from the replay buffer to update the non-parametric values
    while supplying counter-factual values with the parametric model.

    Args:
      trajectories: Current OpenSpiel game state.
    """
    # Calculate non-parametric values over the trajectories.
    # Iterate backward through trajectories
    for t in range(len(trajectories) - 1, 0, -1):
      elem = trajectories[t][1]
      s_tp1 = tuple(elem.next_info_state)
      s_t = tuple(elem.info_state)
      a_t = elem.action
      r_t = elem.reward
      legal_actions = elem.legal_actions_mask
      if t < len(trajectories) - 1:
        for action in range(len(legal_actions)):
          if not legal_actions[action]:
            continue
          if action == elem.action:
            self._q_np[s_t][a_t] = (r_t + self._discount * self._v_np[s_tp1])
          else:
            q_values_parametric = self._session.run(
                self._agent.q_values,
                feed_dict={
                    self._agent.info_state_ph:
                        np.expand_dims(elem.info_state, axis=0)
                })
            self._q_np[s_t][a_t] = q_values_parametric[0][action]

      # Set V(s_t)
      if t == len(trajectories) - 1:
        # Sample from the parametric model.
        q_values_parametric = self._session.run(
            self._agent.q_values,
            feed_dict={
                self._agent.info_state_ph:
                    np.expand_dims(elem.info_state, axis=0)
            })
        self._v_np[s_t] = np.max(q_values_parametric)
      else:
        self._v_np[s_t] = max(self._q_np[s_t])

  def _epsilon_greedy(self, q_values, legal_actions, epsilon):
    """Returns a valid epsilon-greedy action and valid action probs.

    Action probabilities are given by a softmax over legal q-values.

    Args:
      q_values: list of Q-values by action.
      legal_actions: list of legal actions at `info_state`.
      epsilon: float, probability of taking an exploratory action.

    Returns:
      A valid epsilon-greedy action and valid action probabilities.
    """
    probs = np.zeros(self._num_actions)
    q_values = np.array(q_values)
    if np.random.rand() < epsilon:
      action = np.random.choice(legal_actions)
      probs[legal_actions] = 1.0 / len(legal_actions)
    else:
      legal_q_values = q_values[legal_actions]
      action = legal_actions[np.argmax(legal_q_values)]
      # Reduce max_q for numerical stability. Result is the same.
      max_q = np.max(legal_q_values)
      e_x = np.exp(legal_q_values - max_q)
      probs[legal_actions] = e_x / e_x.sum(axis=0)
    return action, probs

  def _get_epsilon(self, step_counter, is_evaluation):
    """Returns the evaluation or decayed epsilon value."""
    if is_evaluation:
      return 0.0
    decay_steps = min(step_counter, self._epsilon_decay_duration)
    decayed_epsilon = (
        self._epsilon_end + (self._epsilon_start - self._epsilon_end) *
        (1 - decay_steps / self._epsilon_decay_duration))
    return decayed_epsilon

  def action_probabilities(self, state):
    """Returns action probabilites dict for a single batch."""
    # TODO(author3, author6): Refactor this to expect pre-normalized form.
    if hasattr(state, "information_state_tensor"):
      state_rep = tuple(state.information_state_tensor(self.player_id))
    elif hasattr(state, "observation_tensor"):
      state_rep = tuple(state.observation_tensor(self.player_id))
    else:
      raise AttributeError("Unable to extract normalized state vector.")
    legal_actions = state.legal_actions(self.player_id)
    if legal_actions:
      _, probs = self._epsilon_greedy(
          self._q_eva[state_rep], legal_actions, epsilon=0.0)
      return {a: probs[a] for a in range(self._num_actions)}
    else:
      raise ValueError("Node has no legal actions to take.")



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Quantal Response Equilibrium (QRE) Stochastic Approximate Nash Solver."""

from absl import logging  # pylint:disable=unused-import

import numpy as np
from scipy import special

from open_spiel.python.algorithms.adidas_utils.helpers import simplex
from open_spiel.python.algorithms.adidas_utils.helpers.symmetric import exploitability as exp


class Solver(object):
  """QRE Solver."""

  def __init__(self, temperature=0., proj_grad=True, euclidean=False,
               cheap=False, lrs=(1e-2, 1e-1), vr=True, rnd_init=False,
               seed=None, **kwargs):
    """Ctor."""
    del kwargs
    if temperature < 0.:
      raise ValueError("temperature must be non-negative")
    self.num_players = None
    self.temperature = temperature
    self.proj_grad = proj_grad
    self.cheap = cheap
    self.vr = vr
    self.pm_vr = None
    self.rnd_init = rnd_init
    self.lrs = lrs
    self.has_aux = True
    self.aux_errors = []

    self.euclidean = euclidean
    if euclidean:
      self.update = self.euc_descent_step
    else:
      self.update = self.mirror_descent_step

    self.seed = seed
    self.random = np.random.RandomState(seed)

  def init_vars(self, num_strats, num_players):
    """Initialize solver parameters."""
    self.num_players = num_players
    if self.rnd_init:
      init_dist = self.random.rand(num_strats)
    else:
      init_dist = np.ones(num_strats)
    init_dist /= init_dist.sum()
    init_y = np.zeros(num_strats)
    if self.cheap and self.vr:
      self.pm_vr = np.zeros((num_strats, num_strats))
    return (init_dist, init_y)

  def record_aux_errors(self, grads):
    """Record errors for the auxiliary variables."""
    grad_y = grads[1]
    self.aux_errors.append([np.linalg.norm(grad_y)])

  def compute_gradients(self, params, payoff_matrices):
    """Compute and return gradients (and exploitabilities) for all parameters.

    Args:
      params: tuple of params (dist, y), see qre.gradients
      payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    Returns:
      tuple of gradients (grad_dist, grad_y, grad_z), see qre.gradients
      unregularized exploitability (stochastic estimate)
      tsallis regularized exploitability (stochastic estimate)
    """
    if self.cheap and self.vr:
      grads, pm_vr, exp_sto, exp_solver_sto = cheap_gradients_vr(
          self.random, *params, payoff_matrices, self.num_players, self.pm_vr,
          self.temperature, self.proj_grad,)
      self.pm_vr = pm_vr
      return grads, exp_sto, exp_solver_sto
    elif self.cheap and not self.vr:
      return cheap_gradients(self.random, *params, payoff_matrices,
                             self.num_players, self.temperature, self.proj_grad)
    else:
      return gradients(*params, payoff_matrices, self.num_players,
                       self.temperature, self.proj_grad)

  def exploitability(self, params, payoff_matrices):
    """Compute and return shannon entropy regularized exploitability.

    Args:
      params: tuple of params (dist, y), see qre.gradients
      payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    Returns:
      float, exploitability of current dist
    """
    return exp.qre_exploitability(params, payoff_matrices, self.temperature)

  def euc_descent_step(self, params, grads, t):
    """Projected gradient descent on exploitability using Euclidean projection.

    Args:
      params: tuple of variables to be updated (dist, y)
      grads: tuple of variable gradients (grad_dist, grad_y)
      t: int, solver iteration
    Returns:
      new_params: tuple of update params (new_dist, new_y)
    """
    lr_dist, lr_y = self.lrs
    new_params = [params[0] - lr_dist * grads[0]]
    lr_y = np.clip(1 / float(t + 1), lr_y, np.inf)
    new_params += [params[1] - lr_y * grads[1]]
    new_params = euc_project(*new_params)
    return new_params

  def mirror_descent_step(self, params, grads, t):
    """Entropic mirror descent on exploitability.

    Args:
      params: tuple of variables to be updated (dist, y)
      grads: tuple of variable gradients (grad_dist, grad_y)
      t: int, solver iteration
    Returns:
      new_params: tuple of update params (new_dist, new_y)
    """
    lr_dist, lr_y = self.lrs
    new_params = [np.log(np.clip(params[0], 0, np.inf)) - lr_dist * grads[0]]
    lr_y = np.clip(1 / float(t + 1), lr_y, np.inf)
    new_params += [params[1] - lr_y * grads[1]]
    new_params = mirror_project(*new_params)
    return new_params


def gradients(dist, y, payoff_matrices, num_players, temperature=0.,
              proj_grad=True):
  """Computes exploitablity gradient and aux variable gradients.

  Args:
    dist: 1-d np.array, current estimate of nash distribution
    y: 1-d np.array (same shape as dist), current estimate of payoff gradient
    payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    num_players: int, number of players, in case payoff_matrices is abbreviated
    temperature: non-negative float, default 0.
    proj_grad: bool, if True, projects dist gradient onto simplex
  Returns:
    gradient of exploitability w.r.t. (dist, y) as tuple
    unregularized exploitability (stochastic estimate)
    tsallis regularized exploitability (stochastic estimate)
  """
  nabla = payoff_matrices[0].dot(dist)
  y = nabla
  if temperature > 0:
    br = special.softmax(y / temperature)
    br_mat = (np.diag(br) - np.outer(br, br)) / temperature
    br_policy_gradient = nabla - temperature * (np.log(br) + 1)
  else:
    power = np.inf
    s = np.linalg.norm(y, ord=power)
    br = np.zeros_like(dist)
    maxima = (y == s)
    br[maxima] = 1. / maxima.sum()
    br_mat = np.zeros((br.size, br.size))
    br_policy_gradient = np.zeros_like(br)

  unreg_exp = np.max(y) - y.dot(dist)
  entr_br = temperature * special.entr(br).sum()
  entr_dist = temperature * special.entr(dist).sum()
  reg_exp = y.dot(br - dist) + entr_br - entr_dist

  policy_gradient = nabla
  if temperature > 0:
    policy_gradient -= temperature * (np.log(dist) + 1)
  other_player_fx = (br - dist) + br_mat.dot(br_policy_gradient)

  other_player_fx_translated = payoff_matrices[1].dot(other_player_fx)
  grad_dist = -policy_gradient + (num_players - 1) * other_player_fx_translated
  if proj_grad:
    grad_dist = simplex.project_grad(grad_dist)
  grad_y = y - nabla

  return (grad_dist, grad_y), unreg_exp, reg_exp


def cheap_gradients(random, dist, y, payoff_matrices, num_players,
                    temperature=0., proj_grad=True):
  """Computes exploitablity gradient and aux variable gradients with samples.

  This implementation takes payoff_matrices as input so technically uses O(d^2)
  compute but only a single column of payoff_matrices is used to perform the
  update so can be re-implemented in O(d) if needed.

  Args:
    random: random number generator, np.random.RandomState(seed)
    dist: 1-d np.array, current estimate of nash distribution
    y: 1-d np.array (same shape as dist), current estimate of payoff gradient
    payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    num_players: int, number of players, in case payoff_matrices is abbreviated
    temperature: non-negative float, default 0.
    proj_grad: bool, if True, projects dist gradient onto simplex
  Returns:
    gradient of exploitability w.r.t. (dist, y) as tuple
    unregularized exploitability (stochastic estimate)
    tsallis regularized exploitability (stochastic estimate)
  """
  action_1 = random.choice(dist.size, p=dist)
  nabla = payoff_matrices[0][:, action_1]
  if temperature > 0:
    br = special.softmax(y / temperature)
    br_mat = (np.diag(br) - np.outer(br, br)) / temperature
    br_policy_gradient = nabla - temperature * (np.log(br) + 1)
  else:
    power = np.inf
    s = np.linalg.norm(y, ord=power)
    br = np.zeros_like(dist)
    maxima = (y == s)
    br[maxima] = 1. / maxima.sum()
    br_mat = np.zeros((br.size, br.size))
    br_policy_gradient = np.zeros_like(br)

  unreg_exp = np.max(y) - y.dot(dist)
  entr_br = temperature * special.entr(br).sum()
  entr_dist = temperature * special.entr(dist).sum()
  reg_exp = y.dot(br - dist) + entr_br - entr_dist

  policy_gradient = nabla
  if temperature > 0:
    policy_gradient -= temperature * (np.log(dist) + 1)
  other_player_fx = (br - dist) + br_mat.dot(br_policy_gradient)

  action_u = random.choice(dist.size)  # uniform, ~importance sampling
  other_player_fx = dist.size * other_player_fx[action_u]
  other_player_fx_translated = payoff_matrices[1, :, action_u] * other_player_fx
  grad_dist = -policy_gradient + (num_players - 1) * other_player_fx_translated
  if proj_grad:
    grad_dist = simplex.project_grad(grad_dist)
  grad_y = y - nabla

  return (grad_dist, grad_y), unreg_exp, reg_exp


def cheap_gradients_vr(random, dist, y, payoff_matrices, num_players, pm_vr,
                       temperature=0., proj_grad=True, version=0):
  """Computes exploitablity gradient and aux variable gradients with samples.

  This implementation takes payoff_matrices as input so technically uses O(d^2)
  compute but only a single column of payoff_matrices is used to perform the
  update so can be re-implemented in O(d) if needed.

  Args:
    random: random number generator, np.random.RandomState(seed)
    dist: 1-d np.array, current estimate of nash distribution
    y: 1-d np.array (same shape as dist), current estimate of payoff gradient
    payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    num_players: int, number of players, in case payoff_matrices is abbreviated
    pm_vr: approximate payoff_matrix for variance reduction
    temperature: non-negative float, default 0.
    proj_grad: bool, if True, projects dist gradient onto simplex
    version: int, default 0, two options for variance reduction
  Returns:
    gradient of exploitability w.r.t. (dist, y) as tuple
    unregularized exploitability (stochastic estimate)
    tsallis regularized exploitability (stochastic estimate)
  """
  if pm_vr is None:
    raise ValueError("pm_vr must be np.array of shape (num_strats, num_strats)")
  if (not isinstance(version, int)) or (version < 0) or (version > 1):
    raise ValueError("version must be non-negative int < 2")

  action_1 = random.choice(dist.size, p=dist)
  nabla = payoff_matrices[0][:, action_1]
  if temperature > 0:
    br = special.softmax(y / temperature)
    br_mat = (np.diag(br) - np.outer(br, br)) / temperature
    br_policy_gradient = nabla - temperature * (np.log(br) + 1)
  else:
    power = np.inf
    s = np.linalg.norm(y, ord=power)
    br = np.zeros_like(dist)
    maxima = (y == s)
    br[maxima] = 1. / maxima.sum()
    br_mat = np.zeros((br.size, br.size))
    br_policy_gradient = np.zeros_like(br)

  unreg_exp = np.max(y) - y.dot(dist)
  entr_br = temperature * special.entr(br).sum()
  entr_dist = temperature * special.entr(dist).sum()
  reg_exp = y.dot(br - dist) + entr_br - entr_dist

  policy_gradient = nabla
  if temperature > 0:
    policy_gradient -= temperature * (np.log(dist) + 1)
  other_player_fx = (br - dist) + br_mat.dot(br_policy_gradient)

  if version == 0:
    other_player_fx_translated = pm_vr.dot(other_player_fx)
    action_u = random.choice(dist.size)  # uniform, ~importance sampling
    other_player_fx = other_player_fx[action_u]
    pm_mod = dist.size * (payoff_matrices[1, :, action_u] - pm_vr[:, action_u])
    other_player_fx_translated += pm_mod * other_player_fx
  elif version == 1:
    other_player_fx_translated = np.sum(pm_vr, axis=1)
    action_u = random.choice(dist.size)  # uniform, ~importance sampling
    other_player_fx = other_player_fx[action_u]
    pm_mod = dist.size * payoff_matrices[1, :, action_u]
    r = dist.size * pm_vr[:, action_u]
    other_player_fx_translated += pm_mod * other_player_fx - r

  grad_dist = -policy_gradient + (num_players - 1) * other_player_fx_translated
  if proj_grad:
    grad_dist = simplex.project_grad(grad_dist)
  grad_y = y - nabla

  if version == 0:
    pm_vr[:, action_u] = payoff_matrices[1, :, action_u]
  elif version == 1:
    pm_vr[:, action_u] = payoff_matrices[1, :, action_u] * other_player_fx

  return (grad_dist, grad_y), pm_vr, unreg_exp, reg_exp


def euc_project(dist, y):
  """Project variables onto their feasible sets (euclidean proj for dist).

  Args:
    dist: 1-d np.array, current estimate of nash distribution
    y: 1-d np.array (same shape as dist), current estimate of payoff gradient
  Returns:
    projected variables (dist, y) as tuple
  """
  dist = simplex.euclidean_projection_onto_simplex(dist)
  y = np.clip(y, 0., np.inf)

  return dist, y


def mirror_project(dist, y):
  """Project variables onto their feasible sets (softmax for dist).

  Args:
    dist: 1-d np.array, current estimate of nash distribution
    y: 1-d np.array (same shape as dist), current estimate of payoff gradient
  Returns:
    projected variables (dist, y) as tuple
  """
  dist = special.softmax(dist)
  y = np.clip(y, 0., np.inf)

  return dist, y



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.adidas_utils.nonsymmetric."""

import itertools

from absl import logging  # pylint:disable=unused-import
from absl.testing import absltest
from absl.testing import parameterized

import numpy as np

from scipy.spatial.distance import cosine

from open_spiel.python.algorithms.adidas_utils.helpers import misc

from open_spiel.python.algorithms.adidas_utils.solvers.nonsymmetric import ate
from open_spiel.python.algorithms.adidas_utils.solvers.nonsymmetric import ped
from open_spiel.python.algorithms.adidas_utils.solvers.nonsymmetric import qre


class ExploitabilityDescentTest(parameterized.TestCase):

  @staticmethod
  def numerical_gradient(fun, x, eps=np.sqrt(np.finfo(float).eps)):
    fun_0 = fun(x)
    num_grad = [np.zeros_like(xi) for xi in x]
    x_plus_dx = [np.copy(xi) for xi in x]
    for i in range(len(x)):
      for j in range(len(x[i])):
        x_plus_dx[i][j] = x[i][j] + eps
        num_grad[i][j] = (fun(x_plus_dx) - fun_0) / eps
        x_plus_dx[i][j] = x[i][j]
    return num_grad

  @staticmethod
  def prep_params(dist, pt, num_params):
    params = [dist]
    if num_params > 1:
      num_players = len(dist)
      nabla = [misc.pt_reduce(pt[i], dist, [i]) for i in range(num_players)]
      params += [nabla]  # policy_gradient
    return tuple(params)

  @parameterized.named_parameters(
      ("PED", (ped, False)),
      ("ATE_p=1", (ate, 1., False)),
      ("ATE_p=0.5", (ate, 0.5, False)),
      ("ATE_p=0.1", (ate, 0.1, False)),
      ("ATE_p=0", (ate, 0., False)),
      ("QRE_t=0.0", (qre, 0.0, False)),
      ("QRE_t=0.1", (qre, 0.1, False))
      )
  def test_exploitability_gradient_on_nonsymmetric_three_player_matrix_games(
      self, solver_tuple, trials=100, max_num_strats=3, atol=1e-1, rtol=1e-1,
      seed=1234):
    num_players = 3
    solver = solver_tuple[0].Solver(*solver_tuple[1:])

    random = np.random.RandomState(seed)

    successes = []
    for _ in range(trials):
      num_strats = random.randint(low=2, high=max_num_strats + 1,
                                  size=num_players)
      num_strats = tuple([int(ns) for ns in num_strats])
      payoff_tensor = random.rand(num_players, *num_strats)

      num_params = len(solver.init_vars(num_strats, num_players))
      dirichlet_alpha = [np.ones(num_strats_i) for num_strats_i in num_strats]
      dist = [random.dirichlet(alpha_i) for alpha_i in dirichlet_alpha]
      params = self.prep_params(dist, payoff_tensor, num_params)

      payoff_matrices = {}
      for pi, pj in itertools.combinations(range(num_players), 2):
        key = (pi, pj)
        pt_i = misc.pt_reduce(payoff_tensor[pi], dist, [pi, pj])
        pt_j = misc.pt_reduce(payoff_tensor[pj], dist, [pi, pj])
        payoff_matrices[key] = np.stack((pt_i, pt_j), axis=0)
      grad = solver.compute_gradients(params, payoff_matrices)[0][0]
      grad = np.concatenate(grad) / float(num_players)

      exp = lambda x: solver.exploitability(x, payoff_tensor)  # pylint: disable=cell-var-from-loop
      num_grad = np.concatenate(self.numerical_gradient(exp, dist))

      successes += [np.logical_and(np.allclose(grad, num_grad, rtol, atol),
                                   cosine(grad, num_grad) <= atol)]

    perc = 100 * np.mean(successes)
    logging.info("gradient accuracy success rate out of %d is %f", trials, perc)
    self.assertGreaterEqual(
        perc, 95., "exploitability gradient accuracy is too poor")


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for crowd_modelling_2d."""

from absl.testing import absltest

from open_spiel.python.mfg.games import crowd_modelling_2d


class CrowdModelling2DTest(absltest.TestCase):

  def test_grid_to_forbidden_states(self):
    forbidden_states = crowd_modelling_2d.grid_to_forbidden_states([
        "#####",
        "# # #",
        "#   #",
        "#####",
    ])

    self.assertEqual(
        forbidden_states,
        "[0|0;1|0;2|0;3|0;4|0;0|1;2|1;4|1;0|2;4|2;0|3;1|3;2|3;3|3;4|3]")


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Tic tac toe (noughts and crosses), implemented in Python.

This is a demonstration of implementing a deterministic perfect-information
game in Python.

Python games are significantly slower than C++, but it may still be suitable
for prototyping or for small games.

It is possible to run C++ algorithms on Python-implemented games. This is likely
to have good performance if the algorithm simply extracts a game tree and then
works with that (e.g. CFR algorithms). It is likely to be poor if the algorithm
relies on processing and updating states as it goes, e.g., MCTS.
"""

import numpy as np

from open_spiel.python.observation import IIGObserverForPublicInfoGame
import pyspiel

_NUM_PLAYERS = 2
_NUM_ROWS = 3
_NUM_COLS = 3
_NUM_CELLS = _NUM_ROWS * _NUM_COLS
_GAME_TYPE = pyspiel.GameType(
    short_name="python_tic_tac_toe",
    long_name="Python Tic-Tac-Toe",
    dynamics=pyspiel.GameType.Dynamics.SEQUENTIAL,
    chance_mode=pyspiel.GameType.ChanceMode.DETERMINISTIC,
    information=pyspiel.GameType.Information.PERFECT_INFORMATION,
    utility=pyspiel.GameType.Utility.ZERO_SUM,
    reward_model=pyspiel.GameType.RewardModel.TERMINAL,
    max_num_players=_NUM_PLAYERS,
    min_num_players=_NUM_PLAYERS,
    provides_information_state_string=True,
    provides_information_state_tensor=False,
    provides_observation_string=True,
    provides_observation_tensor=True,
    parameter_specification={})
_GAME_INFO = pyspiel.GameInfo(
    num_distinct_actions=_NUM_CELLS,
    max_chance_outcomes=0,
    num_players=2,
    min_utility=-1.0,
    max_utility=1.0,
    utility_sum=0.0,
    max_game_length=_NUM_CELLS)


class TicTacToeGame(pyspiel.Game):
  """A Python version of the Tic-Tac-Toe game."""

  def __init__(self, params=None):
    super().__init__(_GAME_TYPE, _GAME_INFO, params or dict())

  def new_initial_state(self):
    """Returns a state corresponding to the start of a game."""
    return TicTacToeState(self)

  def make_py_observer(self, iig_obs_type=None, params=None):
    """Returns an object used for observing game state."""
    if ((iig_obs_type is None) or
        (iig_obs_type.public_info and not iig_obs_type.perfect_recall)):
      return BoardObserver(params)
    else:
      return IIGObserverForPublicInfoGame(iig_obs_type, params)


class TicTacToeState(pyspiel.State):
  """A python version of the Tic-Tac-Toe state."""

  def __init__(self, game):
    """Constructor; should only be called by Game.new_initial_state."""
    super().__init__(game)
    self._cur_player = 0
    self._player0_score = 0.0
    self._is_terminal = False
    self.board = np.full((_NUM_ROWS, _NUM_COLS), ".")

  # OpenSpiel (PySpiel) API functions are below. This is the standard set that
  # should be implemented by every perfect-information sequential-move game.

  def current_player(self):
    """Returns id of the next player to move, or TERMINAL if game is over."""
    return pyspiel.PlayerId.TERMINAL if self._is_terminal else self._cur_player

  def _legal_actions(self, player):
    """Returns a list of legal actions, sorted in ascending order."""
    return [a for a in range(_NUM_CELLS) if self.board[_coord(a)] == "."]

  def _apply_action(self, action):
    """Applies the specified action to the state."""
    self.board[_coord(action)] = "x" if self._cur_player == 0 else "o"
    if _line_exists(self.board):
      self._is_terminal = True
      self._player0_score = 1.0 if self._cur_player == 0 else -1.0
    elif all(self.board.ravel() != "."):
      self._is_terminal = True
    else:
      self._cur_player = 1 - self._cur_player

  def _action_to_string(self, player, action):
    """Action -> string."""
    row, col = _coord(action)
    return "{}({},{})".format("x" if player == 0 else "o", row, col)

  def is_terminal(self):
    """Returns True if the game is over."""
    return self._is_terminal

  def returns(self):
    """Total reward for each player over the course of the game so far."""
    return [self._player0_score, -self._player0_score]

  def __str__(self):
    """String for debug purposes. No particular semantics are required."""
    return _board_to_string(self.board)


class BoardObserver:
  """Observer, conforming to the PyObserver interface (see observation.py)."""

  def __init__(self, params):
    """Initializes an empty observation tensor."""
    if params:
      raise ValueError(f"Observation parameters not supported; passed {params}")
    # The observation should contain a 1-D tensor in `self.tensor` and a
    # dictionary of views onto the tensor, which may be of any shape.
    # Here the observation is indexed `(cell state, row, column)`.
    shape = (1 + _NUM_PLAYERS, _NUM_ROWS, _NUM_COLS)
    self.tensor = np.zeros(np.prod(shape), np.float32)
    self.dict = {"observation": np.reshape(self.tensor, shape)}

  def set_from(self, state, player):
    """Updates `tensor` and `dict` to reflect `state` from PoV of `player`."""
    del player
    # We update the observation via the shaped tensor since indexing is more
    # convenient than with the 1-D tensor. Both are views onto the same memory.
    obs = self.dict["observation"]
    obs.fill(0)
    for row in range(_NUM_ROWS):
      for col in range(_NUM_COLS):
        cell_state = ".ox".index(state.board[row, col])
        obs[cell_state, row, col] = 1

  def string_from(self, state, player):
    """Observation of `state` from the PoV of `player`, as a string."""
    del player
    return _board_to_string(state.board)


# Helper functions for game details.


def _line_value(line):
  """Checks a possible line, returning the winning symbol if any."""
  if all(line == "x") or all(line == "o"):
    return line[0]


def _line_exists(board):
  """Checks if a line exists, returns "x" or "o" if so, and None otherwise."""
  return (_line_value(board[0]) or _line_value(board[1]) or
          _line_value(board[2]) or _line_value(board[:, 0]) or
          _line_value(board[:, 1]) or _line_value(board[:, 2]) or
          _line_value(board.diagonal()) or
          _line_value(np.fliplr(board).diagonal()))


def _coord(move):
  """Returns (row, col) from an action id."""
  return (move // _NUM_COLS, move % _NUM_COLS)


def _board_to_string(board):
  """Returns a string representation of the board."""
  return "\n".join("".join(row) for row in board)


# Register the game with the OpenSpiel library

pyspiel.register_game(_GAME_TYPE, TicTacToeGame)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Tests for Python Tic-Tac-Toe."""

import difflib
import os
import pickle

from absl.testing import absltest
import numpy as np
from open_spiel.python.algorithms.get_all_states import get_all_states
from open_spiel.python.games import tic_tac_toe
from open_spiel.python.observation import make_observation
import pyspiel

_DATA_DIR = "open_spiel/integration_tests/playthroughs/"


class TicTacToeTest(absltest.TestCase):

  def test_can_create_game_and_state(self):
    """Checks we can create the game and a state."""
    game = tic_tac_toe.TicTacToeGame()
    state = game.new_initial_state()
    self.assertEqual(str(state), "...\n...\n...")

  def test_random_game(self):
    """Tests basic API functions."""
    # This is here mostly to show the API by example.
    # More serious simulation tests are done in python/tests/games_sim_test.py
    # and in test_game_from_cc (below), both of which test the conformance to
    # the API thoroughly.
    game = tic_tac_toe.TicTacToeGame()
    state = game.new_initial_state()
    while not state.is_terminal():
      print(state)
      cur_player = state.current_player()
      legal_actions = state.legal_actions()
      action = np.random.choice(legal_actions)
      print("Player {} chooses action {}".format(cur_player, action))
      state.apply_action(action)
    print(state)
    print("Returns: {}".format(state.returns()))

  def test_game_from_cc(self):
    """Runs our standard game tests, checking API consistency."""
    game = pyspiel.load_game("python_tic_tac_toe")
    pyspiel.random_sim_test(game, num_sims=10, serialize=False, verbose=True)

  def test_playthoughs_consistent(self):
    """Checks the saved C++ and Python playthroughs are the same."""
    test_srcdir = os.environ.get("TEST_SRCDIR", "")
    path = os.path.join(test_srcdir, _DATA_DIR)
    cc_playthrough = os.path.join(path, "tic_tac_toe.txt")
    py_playthrough = os.path.join(path, "python_tic_tac_toe.txt")
    with open(cc_playthrough, encoding="utf-8") as cc:
      with open(py_playthrough, encoding="utf-8") as py:
        diffs = difflib.ndiff(list(cc), list(py))
    diffs = {d for d in diffs if d and d[0] in {"+", "-"}}
    self.assertEqual(
        diffs, {
            "- game: tic_tac_toe\n",
            "+ game: python_tic_tac_toe\n",
            '- GameType.long_name = "Tic Tac Toe"\n',
            '+ GameType.long_name = "Python Tic-Tac-Toe"\n',
            '- GameType.short_name = "tic_tac_toe"\n',
            '+ GameType.short_name = "python_tic_tac_toe"\n',
            '- ToString() = "tic_tac_toe()"\n',
            '+ ToString() = "python_tic_tac_toe()"\n',
            "- CurrentPlayer() = -4\n",
            "+ CurrentPlayer() = PlayerId.TERMINAL\n",
            "- Returns() = [0, 0]\n",
            "+ Returns() = [0, -0]\n",
        })

  def test_observation_tensors_same(self):
    """Checks observation tensor is the same from C++ and from Python."""
    game = pyspiel.load_game("python_tic_tac_toe")
    state = game.new_initial_state()
    for a in [4, 5, 2, 3]:
      state.apply_action(a)
    py_obs = make_observation(game)
    py_obs.set_from(state, state.current_player())
    cc_obs = state.observation_tensor()
    np.testing.assert_array_equal(py_obs.tensor, cc_obs)

  def test_pickle(self):
    """Checks pickling and unpickling of game and state."""
    game = pyspiel.load_game("python_tic_tac_toe")
    pickled_game = pickle.dumps(game)
    unpickled_game = pickle.loads(pickled_game)
    self.assertEqual(str(game), str(unpickled_game))
    state = game.new_initial_state()
    for a in [4, 2, 3, 7]:
      state.apply_action(a)
    ser_str = pyspiel.serialize_game_and_state(game, state)
    new_game, new_state = pyspiel.deserialize_game_and_state(ser_str)
    self.assertEqual(str(game), str(new_game))
    self.assertEqual(str(state), str(new_state))
    pickled_state = pickle.dumps(state)
    unpickled_state = pickle.loads(pickled_state)
    self.assertEqual(str(state), str(unpickled_state))

  def test_cloned_state_matches_original_state(self):
    """Check we can clone states successfully."""
    game = tic_tac_toe.TicTacToeGame()
    state = game.new_initial_state()
    state.apply_action(1)
    state.apply_action(2)
    clone = state.clone()

    self.assertEqual(state.history(), clone.history())
    self.assertEqual(state.num_players(), clone.num_players())
    self.assertEqual(state.move_number(), clone.move_number())
    self.assertEqual(state.num_distinct_actions(), clone.num_distinct_actions())

    self.assertEqual(state._cur_player, clone._cur_player)
    self.assertEqual(state._player0_score, clone._player0_score)
    self.assertEqual(state._is_terminal, clone._is_terminal)
    np.testing.assert_array_equal(state.board, clone.board)

  def test_consistent(self):
    """Checks the Python and C++ game implementations are the same."""
    py_game = pyspiel.load_game("python_tic_tac_toe")
    cc_game = pyspiel.load_game("tic_tac_toe")
    py_obs = make_observation(py_game)
    cc_obs = make_observation(cc_game)
    py_states = get_all_states(py_game, to_string=str)
    cc_states = get_all_states(cc_game, to_string=str)
    self.assertCountEqual(list(cc_states), list(py_states))
    for key, cc_state in cc_states.items():
      py_state = py_states[key]
      np.testing.assert_array_equal(py_state.history(), cc_state.history())
      np.testing.assert_array_equal(py_state.returns(), cc_state.returns())
      py_obs.set_from(py_state, 0)
      cc_obs.set_from(cc_state, 0)
      np.testing.assert_array_equal(py_obs.tensor, cc_obs.tensor)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An MCTS Evaluator for an AlphaZero model."""

import numpy as np

from open_spiel.python.algorithms import mcts
import pyspiel
from open_spiel.python.utils import lru_cache


class AlphaZeroEvaluator(mcts.Evaluator):
  """An AlphaZero MCTS Evaluator."""

  def __init__(self, game, model, cache_size=2**16):
    """An AlphaZero MCTS Evaluator."""
    if game.num_players() != 2:
      raise ValueError("Game must be for two players.")
    game_type = game.get_type()
    if game_type.reward_model != pyspiel.GameType.RewardModel.TERMINAL:
      raise ValueError("Game must have terminal rewards.")
    if game_type.dynamics != pyspiel.GameType.Dynamics.SEQUENTIAL:
      raise ValueError("Game must have sequential turns.")

    self._model = model
    self._cache = lru_cache.LRUCache(cache_size)

  def cache_info(self):
    return self._cache.info()

  def clear_cache(self):
    self._cache.clear()

  def _inference(self, state):
    # Make a singleton batch
    obs = np.expand_dims(state.observation_tensor(), 0)
    mask = np.expand_dims(state.legal_actions_mask(), 0)

    # ndarray isn't hashable
    cache_key = obs.tobytes() + mask.tobytes()

    value, policy = self._cache.make(
        cache_key, lambda: self._model.inference(obs, mask))

    return value[0, 0], policy[0]  # Unpack batch

  def evaluate(self, state):
    """Returns a value for the given state."""
    value, _ = self._inference(state)
    return np.array([value, -value])

  def prior(self, state):
    if state.is_chance_node():
      return state.chance_outcomes()
    else:
      # Returns the probabilities for all actions.
      _, policy = self._inference(state)
      return [(action, policy[action]) for action in state.legal_actions()]



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for adidas."""

from absl.testing import absltest

import numpy as np

from open_spiel.python.algorithms import adidas

from open_spiel.python.algorithms.adidas_utils.games.big import ElFarol
from open_spiel.python.algorithms.adidas_utils.games.small import MatrixGame
from open_spiel.python.algorithms.adidas_utils.solvers.symmetric import qre_anneal as qre_anneal_sym


class AdidasTest(absltest.TestCase):

  def test_adidas_on_prisoners_dilemma(self):
    """Tests ADIDAS on a 2-player prisoner's dilemma game."""
    # pylint:disable=bad-whitespace
    pt_r = np.array([[-1, -3],
                     [0,  -2]])
    # pylint:enable=bad-whitespace
    # shift tensor to ensure positivity required if run adidas w/ Tsallis entrpy
    pt_r -= pt_r.min()
    pt_c = pt_r.T  # symmetric game
    pt = np.stack((pt_r, pt_c), axis=0).astype(float)
    pt /= pt.max()  # arbitrary design choice to upper bound entries to 1
    game = MatrixGame(pt, seed=0)
    # for games with more than 2 players, see adidas_utils/games/big.py
    solver = qre_anneal_sym.Solver(temperature=100,
                                   proj_grad=False, euclidean=True,
                                   lrs=(1e-4, 1e-4), exp_thresh=0.01,
                                   rnd_init=True, seed=0)
    # note we set rnd_init to True which initializes adidas' initial
    # approximation to nash to a random point on the simplex. if rnd_init is
    # False, adidas is initialized to uniform which is the Nash equilibrium
    # of the prisoner's dilemma, in which case adidas trivially solves this
    # game in 0 iterations.
    lle = adidas.ADIDAS(seed=0)
    lle.approximate_nash(game, solver, sym=True, num_iterations=1,
                         num_samples=1, num_eval_samples=int(1e5),
                         approx_eval=True, exact_eval=True,
                         avg_trajectory=False)
    self.assertLess(lle.results['exps_exact'][-1], 0.2)

  def test_adidas_on_elfarol(self):
    """Test ADIDAS on a 10-player, symmetric El Farol bar game."""
    game = ElFarol(n=10, c=0.7)
    solver = qre_anneal_sym.Solver(temperature=100,
                                   proj_grad=False, euclidean=False,
                                   lrs=(1e-4, 1e-2), exp_thresh=0.01,
                                   seed=0)
    lle = adidas.ADIDAS(seed=0)
    lle.approximate_nash(game, solver, sym=True, num_iterations=1,
                         num_samples=np.inf, num_eval_samples=int(1e5),
                         approx_eval=True, exact_eval=True,
                         avg_trajectory=False)
    self.assertLess(lle.results['exps_exact'][-1], 0.5)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3.
r"""Policy Gradient based agents implemented in PyTorch.

This class is composed of three policy gradient (PG) algorithms:

- Q-based Policy Gradient (QPG): an "all-actions" advantage actor-critic
algorithm differing from A2C in that all action values are used to estimate the
policy gradient (as opposed to only using the action taken into account):

    baseline = \sum_a pi_a * Q_a
    loss = - \sum_a pi_a * (Q_a - baseline)

where (Q_a - baseline) is the usual advantage. QPG is also known as Mean
Actor-Critic (https://arxiv.org/abs/1709.00503).


- Regret policy gradient (RPG): a PG algorithm inspired by counterfactual regret
minimization (CFR). Unlike standard actor-critic methods (e.g. A2C), the loss is
defined purely in terms of thresholded regrets as follows:

    baseline = \sum_a pi_a * Q_a
    loss = regret = \sum_a relu(Q_a - baseline)

where gradients only flow through the action value (Q_a) part and are blocked on
the baseline part (which is trained separately by usual MSE loss).
The lack of negative sign in the front of the loss represents a switch from
gradient ascent on the score to descent on the loss.


- Regret Matching Policy Gradient (RMPG): inspired by regret-matching, the
policy gradient is by weighted by the thresholded regret:

    baseline = \sum_a pi_a * Q_a
    loss = - \sum_a pi_a * relu(Q_a - baseline)


These algorithms were published in NeurIPS 2018. Paper title: "Actor-Critic
Policy Optimization in Partially Observable Multiagent Environment", the paper
is available at: https://arxiv.org/abs/1810.09026.

- Advantage Actor Critic (A2C): The popular advantage actor critic (A2C)
algorithm. The algorithm uses the baseline (Value function) as a control variate
to reduce variance of the policy gradient. The loss is only computed for the
actions actually taken in the episode as opposed to a loss computed for all
actions in the variants above.

  advantages = returns - baseline
  loss = -log(pi_a) * advantages

The algorithm can be found in the textbook:
https://incompleteideas.net/book/RLbook2018.pdf under the chapter on
`Policy Gradients`.

See  open_spiel/python/pytorch/losses/rl_losses_test.py for an example of the
loss computation.
"""

import collections
import os
from absl import logging
import numpy as np
import torch
from torch import nn
from torch import optim
import torch.nn.functional as F

from open_spiel.python import rl_agent
from open_spiel.python.pytorch.dqn import SonnetLinear
from open_spiel.python.pytorch.losses import rl_losses

Transition = collections.namedtuple(
    "Transition", "info_state action reward discount legal_actions_mask")


class MLPTorso(nn.Module):
  """A specialized half-MLP module when constructing multiple heads.

  Note that every layer includes a ReLU non-linearity activation.
  """

  def __init__(self, input_size, hidden_sizes):
    """Create the MLPTorso.

    Args:
      input_size: (int) number of inputs
      hidden_sizes: (list) sizes (number of units) of each hidden layer
    """

    super(MLPTorso, self).__init__()
    self._layers = []
    # Hidden layers
    for size in hidden_sizes:
      self._layers.append(SonnetLinear(in_size=input_size, out_size=size))
      input_size = size

    self.model = nn.ModuleList(self._layers)

  def forward(self, x):
    for layer in self.model:
      x = layer(x)
    return x


class PolicyGradient(rl_agent.AbstractAgent):
  """RPG Agent implementation in PyTorch.

  See open_spiel/python/examples/single_agent_catch.py for an usage example.
  """

  def __init__(self,
               player_id,
               info_state_size,
               num_actions,
               loss_str="a2c",
               loss_class=None,
               hidden_layers_sizes=(128,),
               batch_size=16,
               critic_learning_rate=0.01,
               pi_learning_rate=0.001,
               entropy_cost=0.01,
               num_critic_before_pi=8,
               additional_discount_factor=1.0,
               max_global_gradient_norm=None,
               optimizer_str="sgd"):
    """Initialize the PolicyGradient agent.

    Args:
      player_id: int, player identifier. Usually its position in the game.
      info_state_size: int, info_state vector size.
      num_actions: int, number of actions per info state.
      loss_str: string or None. If string, must be one of ["rpg", "qpg", "rm",
        "a2c", "neurd"] and defined in `_get_loss_class`. If None, a loss class
        must be passed through `loss_class`. Defaults to "a2c".
      loss_class: Class or None. If Class, it must define the policy gradient
        loss. If None a loss class in a string format must be passed through
        `loss_str`. Defaults to None.
      hidden_layers_sizes: iterable, defines the neural network layers. Defaults
          to (128,), which produces a NN: [INPUT] -> [128] -> ReLU -> [OUTPUT].
      batch_size: int, batch size to use for Q and Pi learning. Defaults to 128.
      critic_learning_rate: float, learning rate used for Critic (Q or V).
        Defaults to 0.001.
      pi_learning_rate: float, learning rate used for Pi. Defaults to 0.001.
      entropy_cost: float, entropy cost used to multiply the entropy loss. Can
        be set to None to skip entropy computation. Defaults to 0.001.
      num_critic_before_pi: int, number of Critic (Q or V) updates before each
        Pi update. Defaults to 8 (every 8th critic learning step, Pi also
        learns).
      additional_discount_factor: float, additional discount to compute returns.
        Defaults to 1.0, in which case, no extra discount is applied.  None that
        users must provide *only one of* `loss_str` or `loss_class`.
      max_global_gradient_norm: float or None, maximum global norm of a gradient
        to which the gradient is shrunk if its value is larger.
      optimizer_str: String defining which optimizer to use. Supported values
        are {sgd, adam}
    """
    assert bool(loss_str) ^ bool(loss_class), "Please provide only one option."
    self._kwargs = locals()
    loss_class = loss_class if loss_class else self._get_loss_class(loss_str)
    self._loss_class = loss_class

    self.player_id = player_id
    self._num_actions = num_actions
    self._layer_sizes = hidden_layers_sizes
    self._batch_size = batch_size
    self._extra_discount = additional_discount_factor
    self._num_critic_before_pi = num_critic_before_pi
    self._max_global_gradient_norm = max_global_gradient_norm

    self._episode_data = []
    self._dataset = collections.defaultdict(list)
    self._prev_time_step = None
    self._prev_action = None

    # Step counters
    self._step_counter = 0
    self._episode_counter = 0
    self._num_learn_steps = 0

    # Keep track of the last training loss achieved in an update step.
    self._last_loss_value = None

    # Network
    # activate final as we plug logit and qvalue heads afterwards.
    self._net_torso = MLPTorso(info_state_size, self._layer_sizes)
    torso_out_size = self._layer_sizes[-1]
    self._policy_logits_layer = SonnetLinear(
        torso_out_size, self._num_actions, activate_relu=False)
    # Do not remove policy_logits_network. Even if it's not used directly here,
    # other code outside this file refers to it.
    self.policy_logits_network = nn.Sequential(self._net_torso,
                                               self._policy_logits_layer)

    self._savers = []

    # Add baseline (V) head for A2C (or Q-head for QPG / RPG / RMPG / NeuRD)
    if optimizer_str == "adam":
      self._critic_optimizer = optim.Adam
    elif optimizer_str == "sgd":
      self._critic_optimizer = optim.SGD
    else:
      raise ValueError("Not implemented, choose from 'adam' and 'sgd'.")

    if loss_class.__name__ == "BatchA2CLoss":
      self._baseline_layer = SonnetLinear(
          torso_out_size, 1, activate_relu=False)
      self._critic_network = nn.Sequential(self._net_torso,
                                           self._baseline_layer)
    else:
      self._q_values_layer = SonnetLinear(
          torso_out_size, self._num_actions, activate_relu=False)
      self._critic_network = nn.Sequential(self._net_torso,
                                           self._q_values_layer)

    self._critic_optimizer = self._critic_optimizer(
        self._critic_network.parameters(), lr=critic_learning_rate)

    # Pi loss
    self.pg_class = loss_class(entropy_cost=entropy_cost)
    self._pi_network = nn.Sequential(self._net_torso, self._policy_logits_layer)
    if optimizer_str == "adam":
      self._pi_optimizer = optim.Adam(
          self._pi_network.parameters(), lr=pi_learning_rate)
    elif optimizer_str == "sgd":
      self._pi_optimizer = optim.SGD(
          self._pi_network.parameters(), lr=pi_learning_rate)

    self._loss_str = loss_str

  def _get_loss_class(self, loss_str):
    if loss_str == "rpg":
      return rl_losses.BatchRPGLoss
    elif loss_str == "qpg":
      return rl_losses.BatchQPGLoss
    elif loss_str == "rm":
      return rl_losses.BatchRMLoss
    elif loss_str == "a2c":
      return rl_losses.BatchA2CLoss
    elif loss_str == "neurd":
      return rl_losses.BatchNeuRDLoss

  def minimize_with_clipping(self, model, optimizer, loss):
    optimizer.zero_grad()
    loss.backward()
    if self._max_global_gradient_norm is not None:
      nn.utils.clip_grad_norm_(model.parameters(),
                               self._max_global_gradient_norm)
    optimizer.step()

  def _act(self, info_state, legal_actions):
    # Make a singleton batch for NN compatibility: [1, info_state_size]
    info_state = torch.Tensor(np.reshape(info_state, [1, -1]))
    torso_out = self._net_torso(info_state)
    self._policy_logits = self._policy_logits_layer(torso_out)
    policy_probs = F.softmax(self._policy_logits, dim=1).detach()

    # Remove illegal actions, re-normalize probs
    probs = np.zeros(self._num_actions)
    probs[legal_actions] = policy_probs[0][legal_actions]
    if sum(probs) != 0:
      probs /= sum(probs)
    else:
      probs[legal_actions] = 1 / len(legal_actions)
    action = np.random.choice(len(probs), p=probs)
    return action, probs

  def step(self, time_step, is_evaluation=False):
    """Returns the action to be taken and updates the network if needed.

    Args:
      time_step: an instance of rl_environment.TimeStep.
      is_evaluation: bool, whether this is a training or evaluation call.

    Returns:
      A `rl_agent.StepOutput` containing the action probs and chosen action.
    """
    # Act step: don't act at terminal info states or if its not our turn.
    if (not time_step.last()) and (
        time_step.is_simultaneous_move() or
        self.player_id == time_step.current_player()):
      info_state = time_step.observations["info_state"][self.player_id]
      legal_actions = time_step.observations["legal_actions"][self.player_id]
      action, probs = self._act(info_state, legal_actions)
    else:
      action = None
      probs = []

    if not is_evaluation:
      self._step_counter += 1

      # Add data points to current episode buffer.
      if self._prev_time_step:
        self._add_transition(time_step)

      # Episode done, add to dataset and maybe learn.
      if time_step.last():
        self._add_episode_data_to_dataset()
        self._episode_counter += 1

        if len(self._dataset["returns"]) >= self._batch_size:
          self._critic_update()
          self._num_learn_steps += 1
          if self._num_learn_steps % self._num_critic_before_pi == 0:
            self._pi_update()
          self._dataset = collections.defaultdict(list)

        self._prev_time_step = None
        self._prev_action = None
        return
      else:
        self._prev_time_step = time_step
        self._prev_action = action

    return rl_agent.StepOutput(action=action, probs=probs)

  def _full_checkpoint_name(self, checkpoint_dir, name):
    checkpoint_filename = "_".join(
        [self._loss_str, name, "pid" + str(self.player_id)])
    return os.path.join(checkpoint_dir, checkpoint_filename)

  def _latest_checkpoint_filename(self, name):
    checkpoint_filename = "_".join(
        [self._loss_str, name, "pid" + str(self.player_id)])
    return checkpoint_filename + "_latest"

  def save(self, checkpoint_dir):
    for name, model in self._savers:
      path = self._full_checkpoint_name(checkpoint_dir, name)
      torch.save(model.state_dict(), path)
      logging.info("Saved to path: %s", path)

  def has_checkpoint(self, checkpoint_dir):
    for name, _ in self._savers:
      path = self._full_checkpoint_name(checkpoint_dir, name)
      if os.path.exists(path):
        return True
    return False

  def restore(self, checkpoint_dir):
    for name, model in self._savers:
      full_checkpoint_dir = self._full_checkpoint_name(checkpoint_dir, name)
      logging.info("Restoring checkpoint: %s", full_checkpoint_dir)
      model.load_state_dict(torch.load(full_checkpoint_dir))

  @property
  def loss(self):
    return (self._last_critic_loss_value, self._last_pi_loss_value)

  def _add_episode_data_to_dataset(self):
    """Add episode data to the buffer."""
    info_states = [data.info_state for data in self._episode_data]
    rewards = [data.reward for data in self._episode_data]
    discount = [data.discount for data in self._episode_data]
    actions = [data.action for data in self._episode_data]

    # Calculate returns
    returns = np.array(rewards)
    for idx in reversed(range(len(rewards[:-1]))):
      returns[idx] = (
          rewards[idx] +
          discount[idx] * returns[idx + 1] * self._extra_discount)

    # Add flattened data points to dataset
    self._dataset["actions"].extend(actions)
    self._dataset["returns"].extend(returns)
    self._dataset["info_states"].extend(info_states)
    self._episode_data = []

  def _add_transition(self, time_step):
    """Adds intra-episode transition to the `_episode_data` buffer.

    Adds the transition from `self._prev_time_step` to `time_step`.

    Args:
      time_step: an instance of rl_environment.TimeStep.
    """
    assert self._prev_time_step is not None
    legal_actions = (
        self._prev_time_step.observations["legal_actions"][self.player_id])
    legal_actions_mask = np.zeros(self._num_actions)
    legal_actions_mask[legal_actions] = 1.0
    transition = Transition(
        info_state=(
            self._prev_time_step.observations["info_state"][self.player_id][:]),
        action=self._prev_action,
        reward=time_step.rewards[self.player_id],
        discount=time_step.discounts[self.player_id],
        legal_actions_mask=legal_actions_mask)

    self._episode_data.append(transition)

  def _critic_update(self):
    """Compute the Critic loss on sampled transitions & perform a critic update.

    Returns:
      The average Critic loss obtained on this batch.
    """
    # TODO(author3): illegal action handling.
    info_state = torch.Tensor(self._dataset["info_states"])
    action = torch.LongTensor(self._dataset["actions"])
    return_ = torch.Tensor(self._dataset["returns"])
    torso_out = self._net_torso(info_state)

    # Critic loss
    # Baseline loss in case of A2C
    if self._loss_class.__name__ == "BatchA2CLoss":
      baseline = torch.squeeze(self._baseline_layer(torso_out), dim=1)
      critic_loss = torch.mean(F.mse_loss(baseline, return_))
      self.minimize_with_clipping(self._baseline_layer, self._critic_optimizer,
                                  critic_loss)
    else:
      # Q-loss otherwise.
      q_values = self._q_values_layer(torso_out)
      action_indices = torch.stack(
          [torch.arange(q_values.shape[0], dtype=torch.long), action], dim=0)
      value_predictions = q_values[list(action_indices)]
      critic_loss = torch.mean(F.mse_loss(value_predictions, return_))
      self.minimize_with_clipping(self._q_values_layer, self._critic_optimizer,
                                  critic_loss)
    self._last_critic_loss_value = critic_loss
    return critic_loss

  def _pi_update(self):
    """Compute the Pi loss on sampled transitions and perform a Pi update.

    Returns:
      The average Pi loss obtained on this batch.
    """
    # TODO(author3): illegal action handling.
    info_state = torch.Tensor(self._dataset["info_states"])
    action = torch.LongTensor(self._dataset["actions"])
    return_ = torch.Tensor(self._dataset["returns"])
    torso_out = self._net_torso(info_state)
    self._policy_logits = self._policy_logits_layer(torso_out)

    if self._loss_class.__name__ == "BatchA2CLoss":
      baseline = torch.squeeze(self._baseline_layer(torso_out), dim=1)
      pi_loss = self.pg_class.loss(
          policy_logits=self._policy_logits,
          baseline=baseline,
          actions=action,
          returns=return_)
      self.minimize_with_clipping(self._policy_logits_layer, self._pi_optimizer,
                                  pi_loss)
    else:
      q_values = self._q_values_layer(torso_out)
      pi_loss = self.pg_class.loss(
          policy_logits=self._policy_logits, action_values=q_values)
      self.minimize_with_clipping(self._policy_logits_layer, self._pi_optimizer,
                                  pi_loss)
    self._last_pi_loss_value = pi_loss
    return pi_loss

  def get_weights(self):
    variables = [m.weight for m in self._net_torso.model]
    variables.append(self._policy_logits_layer.weight)
    if self._loss_class.__name__ == "BatchA2CLoss":
      variables.append(self._baseline_layer.weight)
    else:
      variables.append(self._q_values_layer.weight)
    return variables

  def copy_with_noise(self, sigma=0.0, copy_weights=True):
    """Copies the object and perturbates its network's weights with noise.

    Args:
      sigma: gaussian dropout variance term : Multiplicative noise following
        (1+sigma*epsilon), epsilon standard gaussian variable, multiplies each
        model weight. sigma=0 means no perturbation.
      copy_weights: Boolean determining whether to copy model weights (True) or
        just model hyperparameters.

    Returns:
      Perturbated copy of the model.
    """
    _ = self._kwargs.pop("self", None)
    copied_object = PolicyGradient(**self._kwargs)

    net_torso = getattr(copied_object, "_net_torso")
    policy_logits_layer = getattr(copied_object, "_policy_logits_layer")
    if hasattr(copied_object, "_q_values_layer"):
      q_values_layer = getattr(copied_object, "_q_values_layer")
    if hasattr(copied_object, "_baseline_layer"):
      baseline_layer = getattr(copied_object, "_baseline_layer")

    if copy_weights:
      with torch.no_grad():
        for layer in net_torso.model:
          layer.weight *= (1 + sigma * torch.randn(layer.weight.shape))

        policy_logits_layer.weight *= (
            1 + sigma * torch.randn(policy_logits_layer.weight.shape))

        if hasattr(copied_object, "_q_values_layer"):
          q_values_layer.weight *= (
              1 + sigma * torch.randn(q_values_layer.weight.shape))

        if hasattr(copied_object, "_baseline_layer"):
          baseline_layer.weight *= (
              1 + sigma * torch.randn(baseline_layer.weight.shape))

    return copied_object



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for ResponseGraphUCB."""

import itertools

import matplotlib.pyplot as plt
import numpy as np
import scipy.stats

from open_spiel.python.algorithms import fictitious_play
from open_spiel.python.egt import utils as egt_utils
import pyspiel
from open_spiel.python.utils import file_utils


def get_method_tuple_acronym(method_tuple):
  """Returns pretty acronym for specified ResponseGraphUCB method tuple."""
  if isinstance(method_tuple, tuple):
    acronyms = [get_method_acronym(m) for m in method_tuple]
    return ', '.join(acronyms)
  else:
    return get_method_acronym(method_tuple)


def get_method_tuple_linespecs(method):
  """Gets plot linespecs for the specified ResponseGraphUCB method."""
  sampling_strats = [
      'uniform-exhaustive', 'uniform', 'valence-weighted', 'count-weighted'
  ]
  conf_methods = ['ucb-standard', 'clopper-pearson-ucb']
  method_to_id_map = dict(
      (m, i)
      for i, m in enumerate(itertools.product(sampling_strats, conf_methods)))

  # Create palette
  num_colors = len(method_to_id_map.keys())
  colors = plt.get_cmap('Set1', num_colors).colors

  # Spec out the linestyle
  base_method = (method[0], method[1].replace('-relaxed', '')
                )  # Method name without -relaxed suffix
  linespecs = {
      'color': colors[method_to_id_map[base_method]]
  }  # Use base method for color (ignoring relaxed vs non-relaxed)
  if 'relaxed' in method[1]:  # Use actual method for linestyle
    linespecs['linestyle'] = 'dashed'
  else:
    linespecs['linestyle'] = 'solid'

  return linespecs


def get_method_acronym(method):
  """Gets pretty acronym for specified ResponseGraphUCB method."""
  if method == 'uniform-exhaustive':
    return r'$\mathcal{S}$: UE'
  elif method == 'uniform':
    return r'$\mathcal{S}$: U'
  elif method == 'valence-weighted':
    return r'$\mathcal{S}$: VW'
  elif method == 'count-weighted':
    return r'$\mathcal{S}$: CW'
  elif method == 'ucb-standard':
    return r'$\mathcal{C}(\delta)$: UCB'
  elif method == 'ucb-standard-relaxed':
    return r'$\mathcal{C}(\delta)$: R-UCB'
  elif method == 'clopper-pearson-ucb':
    return r'$\mathcal{C}(\delta)$: CP-UCB'
  elif method == 'clopper-pearson-ucb-relaxed':
    return r'$\mathcal{C}(\delta)$: R-CP-UCB'
  elif method == 'fixedbudget-uniform':
    return r'$\mathcal{S}$: U, $\mathcal{C}(\delta)$: FB'
  else:
    raise ValueError('Unknown sampler method: {}!'.format(method))


def digraph_edge_hamming_dist(g1, g2):
  """Returns number of directed edge mismatches between digraphs g1 and g2."""
  dist = 0
  for e1 in g1.edges:
    if e1 not in g2.edges:
      dist += 1
  return dist


class BernoulliGameSampler(object):
  """A sampler for a game with Bernoulli-distributed payoffs."""

  def __init__(self, strategy_spaces, means, payoff_bounds):
    """Initializes the Bernoulli game sampler.

    Payoffs are automatically scaled to lie between 0 and 1.

    Args:
      strategy_spaces: a list of sizes of player strategy spaces.
      means: 1+num_players dimensional array of mean payoffs.
      payoff_bounds: min/max observable value of payoffs, necessary since one
        may seek Bernoulli-sampling for games with different payoff ranges.
    """
    self.strategy_spaces = strategy_spaces
    self.n_players = len(strategy_spaces)
    self.raw_means = means
    self.payoff_bounds = payoff_bounds
    self.means = self.rescale_payoff(means)

    # Specific to the Bernoulli case. Probas in [0,1], proportional to payoffs
    self.p_max = self.means

  def rescale_payoff(self, payoff):
    """Rescales payoffs to be in [0,1]."""
    # Assumes payoffs are bounded between [-payoff_bound, payoff_bound]
    return (payoff - self.payoff_bounds[0]) / (
        self.payoff_bounds[1] - self.payoff_bounds[0])

  def observe_result(self, strat_profile):
    """Returns empirical payoffs for each agent."""
    outcomes = np.zeros(self.n_players)
    for k in range(self.n_players):
      # compute Bernoulli probabilities
      outcomes[k] = np.random.choice(
          [1, 0],
          p=[self.p_max[k][strat_profile], 1. - self.p_max[k][strat_profile]])
    return outcomes


class ZeroSumBernoulliGameSampler(BernoulliGameSampler):
  """A sampler for a zero-sum game with Bernoulli-distributed payoffs."""

  def __init__(self, strategy_spaces, means, payoff_bounds):
    super(ZeroSumBernoulliGameSampler, self).__init__(strategy_spaces, means,
                                                      payoff_bounds)
    # Verify the game is zero-sum
    assert np.allclose(np.sum(self.means, axis=0), 1.)

  def observe_result(self, strat_profile):
    outcomes = np.zeros(self.n_players)
    win_ix = np.random.choice(
        self.n_players, p=self.means[(slice(None),) + strat_profile])
    outcomes[win_ix] = 1.
    return outcomes


def get_payoffs_bernoulli_game(size=(2, 2, 2)):
  """Gets randomly-generated zero-sum symmetric two-player game."""
  too_close = True
  while too_close:
    M = np.random.uniform(-1, 1, size=size)  # pylint: disable=invalid-name
    M[0, :, :] = 0.5 * (M[0, :, :] - M[0, :, :].T)
    M[1, :, :] = -M[0, :, :]
    if np.abs(M[0, 0, 1]) < 0.1:
      too_close = True
    else:
      too_close = False
  return M


def get_soccer_data():
  """Returns the payoffs and strategy labels for MuJoCo soccer experiments."""
  payoff_file = file_utils.find_file(
      'open_spiel/data/paper_data/response_graph_ucb/soccer.txt', 2)
  payoffs = np.loadtxt(payoff_file)
  return payoffs


def get_kuhn_poker_data(num_players=4, iterations=3):
  """Returns the kuhn poker data for the number of players specified."""
  game = pyspiel.load_game('kuhn_poker', {'players': num_players})
  xfp_solver = fictitious_play.XFPSolver(game, save_oracles=True)
  for _ in range(iterations):
    xfp_solver.iteration()

  # Results are seed-dependent, so show some interesting cases
  if num_players == 2:
    meta_games = xfp_solver.get_empirical_metagame(100, seed=1)
  elif num_players == 3:
    meta_games = xfp_solver.get_empirical_metagame(100, seed=5)
  elif num_players == 4:
    meta_games = xfp_solver.get_empirical_metagame(100, seed=2)

  # Metagame utility matrices for each player
  payoff_tables = []
  for i in range(num_players):
    payoff_tables.append(meta_games[i])
  return payoff_tables


def get_game_for_sampler(game_name):
  """Returns pre-processed game data for ResponseGraphUCB examples."""
  # pylint: disable=invalid-name
  if game_name == 'bernoulli':
    M = get_payoffs_bernoulli_game()
    strategy_spaces = [2, 2]
    G = ZeroSumBernoulliGameSampler(
        strategy_spaces, means=M, payoff_bounds=[-1., 1.])
  elif game_name == 'soccer':
    M = get_soccer_data()
    M = M * 2. - 1  # Convert to zero-sum
    strategy_spaces = np.shape(M)
    M = np.asarray([M, M.T])
    G = ZeroSumBernoulliGameSampler(strategy_spaces, means=M,
                                    payoff_bounds=[np.min(M), np.max(M)])
  elif game_name in ['kuhn_poker_2p', 'kuhn_poker_3p', 'kuhn_poker_4p']:
    if '2p' in game_name:
      num_players = 2
    elif '3p' in game_name:
      num_players = 3
    elif '4p' in game_name:
      num_players = 4
    M = get_kuhn_poker_data(num_players, iterations=2)  # pylint: disable=invalid-name
    strategy_spaces = egt_utils.get_num_strats_per_population(M, False)
    G = BernoulliGameSampler(
        strategy_spaces, means=M, payoff_bounds=[np.min(M), np.max(M)])
  else:
    raise ValueError('Game', game_name, 'not implemented!')
  # pylint: enable=invalid-name
  return G


def plot_timeseries(ax, id_ax, data, xticks, xlabel='', ylabel='',
                    label='', logx=False, logy=False, zorder=10,
                    linespecs=None):
  """Plots timeseries data with error bars."""
  if logx:
    ax[id_ax].set_xscale('log')
  if logy:
    ax[id_ax].set_yscale('log')
  if linespecs:
    kwargs = {'color': linespecs['color']}
  else:
    kwargs = {}

  # Seaborn's bootstrapped confidence intervals were used in the original paper
  se = scipy.stats.sem(data, axis=0)
  ax[id_ax].fill_between(xticks, data.mean(0)-se, data.mean(0)+se,
                         zorder=zorder, alpha=0.2, **kwargs)
  ax[id_ax].plot(xticks, data.mean(0), label=label, zorder=zorder, **kwargs)

  # There may be multiple lines on the current axis, some from previous calls to
  # plot_timeseries, so reference just the latest
  if linespecs:
    ax[id_ax].get_lines()[-1].set_dashes([5, 5])
    ax[id_ax].get_lines()[-1].set_linestyle(linespecs['linestyle'])

  ax[id_ax].set(xlabel=xlabel, ylabel=ylabel)
  ax[id_ax].set_axisbelow(True)
  ax[id_ax].grid(True)
  for _, spine in ax[id_ax].spines.items():
    spine.set_zorder(-1)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wraps third-party bridge bots to make them usable in OpenSpiel.

This code enables OpenSpiel interoperation for bots which implement the BlueChip
bridge protocol. This is widely used, e.g. in the World computer bridge
championships. For a rough outline of the protocol, see:
http://www.bluechipbridge.co.uk/protocol.htm

No formal specification is available. This implementation has been verified
to work correctly with WBridge5.

This bot controls a single player in the game of uncontested bridge bidding. It
chooses its actions by invoking an external bot which plays the full game of
bridge. This means that each time the bot is asked for an action, it sends up to
three actions (forced passes from both opponents, plus partner's most recent
action) to the external bridge bot, and obtains an action in return.

Since we are restricting ourselves to the uncontested bidding game, we have
no support for Doubling, Redoubling, or the play of the cards.
"""

import re

import pyspiel

# Example session:
#
# Recv: Connecting "WBridge5" as ANYPL using protocol version 18
# Send: WEST ("WBridge5") seated
# Recv: WEST ready for teams
# Send: Teams: N/S "silent" E/W "bidders"
# Recv: WEST ready to start
# Send: Start of board
# Recv: WEST ready for deal
# Send: Board number 8. Dealer WEST. Neither vulnerable.
# Recv: WEST ready for cards
# Send: WEST's cards: S A T 9 5. H K 6 5. D Q J 8 7 6. C 7.
# Recv: WEST PASSES
# Recv: WEST ready for  NORTH's bid
# Send: EAST PASSES
# Recv: WEST ready for EAST's bid
# Send: EAST bids 1C
# Recv: WEST ready for  SOUTH's bid

# Template regular expressions for messages we receive
_CONNECT = 'Connecting "(?P<client_name>.*)" as ANYPL using protocol version 18'
_SELF_BID_OR_PASS = "{seat} ((?P<pass>PASSES)|bids (?P<bid>[^ ]*))( Alert.)?"

# Templates for fixed messages we receive
_READY_FOR_TEAMS = "{seat} ready for teams"
_READY_TO_START = "{seat} ready to start"
_READY_FOR_DEAL = "{seat} ready for deal"
_READY_FOR_CARDS = "{seat} ready for cards"
_READY_FOR_BID = "{seat} ready for {other}'s bid"

# Templates for messages we send
_SEATED = '{seat} ("{client_name}") seated'
_TEAMS = 'Teams: N/S "opponents" E/W "bidders"'
_START_BOARD = "start of board"
# The board number is arbitrary, but "8" is consistent with the dealer and
# vulnerability we want (in the standard numbering). See Law 2:
# http://web2.acbl.org/documentlibrary/play/Laws-of-Duplicate-Bridge.pdf
_DEAL = "Board number 8. Dealer WEST. Neither vulnerable."
_CARDS = "{seat}'s cards: {hand}"
_OTHER_PLAYER_PASS = "{player} PASSES"
_OTHER_PLAYER_BID = "{player} bids {bid}"

# BlueChip bridge protocol message constants
_SEATS = ["WEST", "EAST"]
_OPPONENTS = ["NORTH", "SOUTH"]
_TRUMP_SUIT = ["C", "D", "H", "S", "NT"]
_NUMBER_TRUMP_SUITS = len(_TRUMP_SUIT)
_RANKS = ["2", "3", "4", "5", "6", "7", "8", "9", "T", "J", "Q", "K", "A"]

# OpenSpiel constants
_PASS_ACTION = 0


def _string_to_action(call_str):
  """Converts a BlueChip bid string to an OpenSpiel action id (an integer).

  Args:
    call_str: string representing a bid in the BlueChip format, i.e. "[level]
      (as a digit) + [trump suit (S, H, D, C or NT)]", e.g. "1C".

  Returns:
    An integer action id - see `bridge_uncontested_bidding.cc`, functions
    `Denomination` and `Level`.
    0 is reserved for Pass, so bids are in order from 1 upwards: 1 = 1C,
    2 = 1D, etc.
  """
  level = int(call_str[0])
  trumps = _TRUMP_SUIT.index(call_str[1:])
  return (level - 1) * _NUMBER_TRUMP_SUITS + trumps + 1


def _action_to_string(action):
  """Converts OpenSpiel action id (an integer) to a BlueChip bid string.

  Args:
    action: an integer action id corresponding to a bid.

  Returns:
    A string in BlueChip format.

  Inverse of `_string_to_action`. See documentation there.
  """
  level = str((action - 1) // _NUMBER_TRUMP_SUITS + 1)
  trumps = _TRUMP_SUIT[(action - 1) % _NUMBER_TRUMP_SUITS]
  return level + trumps


def _expect_regex(client, regex):
  """Reads a line from the client, parses it using the regular expression."""
  line = client.read_line()
  match = re.match(regex, line)
  if not match:
    raise ValueError("Received '{}' which does not match regex '{}'".format(
        line, regex))
  return match.groupdict()


def _expect(client, expected):
  """Reads a line from the client, checks it matches expected line exactly."""
  line = client.read_line()
  if expected != line:
    raise ValueError("Received '{}' but expected '{}'".format(line, expected))


def _hand_string(state_vec):
  """Returns the hand of the to-play player in the state in BlueChip format."""
  # See UncontestedBiddingState::InformationStateTensor
  # The first 52 elements are whether or not we hold the given card (cards
  # ordered suit-by-suit, in ascending order of rank).
  suits = []
  for suit in reversed(range(4)):
    cards = []
    for rank in reversed(range(13)):
      if state_vec[rank * 4 + suit]:
        cards.append(_RANKS[rank])
    suits.append(_TRUMP_SUIT[suit] + " " + (" ".join(cards) if cards else "-") +
                 ".")
  return " ".join(suits)


def _actions(state_vec):
  """Returns the player actions that have been taken in the game so far."""
  # See UncontestedBiddingState::InformationStateTensor
  # The first 52 elements are the cards held, then two elements for each
  # possible action, specifying which of the two players has taken it (if
  # either player has). Then two elements specifying which player we are.
  actions = state_vec[52:-2]
  return [index // 2 for index, value in enumerate(actions) if value]


def _connect(client, seat, state_vec):
  """Performs the initial handshake with a BlueChip bot."""
  client.start()
  client_name = _expect_regex(client, _CONNECT)["client_name"]
  client.send_line(_SEATED.format(seat=seat, client_name=client_name))
  _expect(client, _READY_FOR_TEAMS.format(seat=seat))
  client.send_line(_TEAMS)
  _expect(client, _READY_TO_START.format(seat=seat))
  client.send_line(_START_BOARD)
  _expect(client, _READY_FOR_DEAL.format(seat=seat))
  client.send_line(_DEAL)
  _expect(client, _READY_FOR_CARDS.format(seat=seat))
  client.send_line(_CARDS.format(seat=seat, hand=_hand_string(state_vec)))


class BlueChipBridgeBot(pyspiel.Bot):
  """An OpenSpiel bot, wrapping a BlueChip bridge bot implementation."""

  def __init__(self, game, player_id, client):
    """Initializes an OpenSpiel `Bot` wrapping a BlueChip-compatible bot.

    Args:
      game: The OpenSpiel game object, should be an instance of
        bridge_uncontested_bidding, without forced actions.
      player_id: The id of the player the bot will act as, 0 = West (dealer), 1
        = East.
      client: The BlueChip bot; must support methods `start`, `read_line`, and
        `send_line`.
    """
    pyspiel.Bot.__init__(self)
    self._game = game
    self._player_id = player_id
    self._client = client
    self._seat = _SEATS[player_id]
    self._partner = _SEATS[1 - player_id]
    self._left_hand_opponent = _OPPONENTS[player_id]
    self._right_hand_opponent = _OPPONENTS[1 - player_id]
    self._connected = False

  def player_id(self):
    return self._player_id

  def restart(self):
    """Indicates that the next step may be from a non-sequential state."""
    self._connected = False

  def restart_at(self, state):
    """Indicates that the next step may be from a non-sequential state."""
    self._connected = False

  def step(self, state):
    """Returns the action and policy for the bot in this state."""
    state_vec = state.information_state_tensor(self.player_id())

    # Connect if necessary.
    if not self._connected:
      _connect(self._client, self._seat, state_vec)
      self._connected = True

    # Get the actions in the game so far.
    actions = _actions(state_vec)

    # Unless this is the first or second action in the game, our
    # left-hand-opponent will have passed since our last turn.
    if len(actions) > 1:
      _expect(
          self._client,
          _READY_FOR_BID.format(
              seat=self._seat, other=self._left_hand_opponent))
      self._client.send_line(
          _OTHER_PLAYER_PASS.format(player=self._left_hand_opponent))

    # Unless there aren't any prior actions, our partner will have bid
    # or passed since our last turn, and so we need to send partner's action
    # to the bot.
    if actions:
      _expect(self._client,
              _READY_FOR_BID.format(seat=self._seat, other=self._partner))
      if actions[-1] == _PASS_ACTION:
        self._client.send_line(_OTHER_PLAYER_PASS.format(player=self._partner))
      else:
        self._client.send_line(
            _OTHER_PLAYER_BID.format(
                player=self._partner, bid=_action_to_string(actions[-1])))

    # Unless there aren't any prior actions, our right-hand-opponent will have
    # passed since our last turn.
    if actions:
      _expect(
          self._client,
          _READY_FOR_BID.format(
              seat=self._seat, other=self._right_hand_opponent))
      self._client.send_line(
          _OTHER_PLAYER_PASS.format(player=self._right_hand_opponent))

    # Get our action from the bot.
    our_action = _expect_regex(self._client,
                               _SELF_BID_OR_PASS.format(seat=self._seat))
    action = 0 if our_action["pass"] else _string_to_action(our_action["bid"])
    return (action, 1.0), action



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""DQN agents trained on an MFG against a crowd following a uniform policy."""

from absl import flags
import jax

from open_spiel.python import policy
from open_spiel.python import rl_agent_policy
from open_spiel.python import rl_environment
from open_spiel.python.jax import dqn
from open_spiel.python.mfg.algorithms import distribution
from open_spiel.python.mfg.algorithms import nash_conv
from open_spiel.python.mfg.algorithms import policy_value
from open_spiel.python.mfg.games import factory
from open_spiel.python.utils import app
from open_spiel.python.utils import metrics

FLAGS = flags.FLAGS

flags.DEFINE_string("game_name", "python_mfg_predator_prey",
                    "Name of the game.")
flags.DEFINE_string(
    "env_setting", None,
    "Name of the game settings. If None, the game name will be used.")
flags.DEFINE_integer("num_train_episodes", int(20e6),
                     "Number of training episodes.")
flags.DEFINE_integer("eval_every", 10000,
                     "Episode frequency at which the agents are evaluated.")
flags.DEFINE_list("hidden_layers_sizes", [
    128,
    128,
], "Number of hidden units in the avg-net and Q-net.")
flags.DEFINE_integer("replay_buffer_capacity", int(2e5),
                     "Size of the replay buffer.")
flags.DEFINE_integer("min_buffer_size_to_learn", 1000,
                     "Number of samples in buffer before learning begins.")
flags.DEFINE_integer("batch_size", 128,
                     "Number of transitions to sample at each learning step.")
flags.DEFINE_integer("learn_every", 64,
                     "Number of steps between learning updates.")
flags.DEFINE_float("rl_learning_rate", 0.01,
                   "Learning rate for inner rl agent.")
flags.DEFINE_string("optimizer_str", "sgd",
                    "Optimizer, choose from 'adam', 'sgd'.")
flags.DEFINE_string("loss_str", "mse",
                    "Loss function, choose from 'mse', 'huber'.")
flags.DEFINE_integer("update_target_network_every", 19200,
                     "Number of steps between DQN target network updates.")
flags.DEFINE_float("discount_factor", 1.0,
                   "Discount factor for future rewards.")
flags.DEFINE_integer("epsilon_decay_duration", int(20e6),
                     "Number of game steps over which epsilon is decayed.")
flags.DEFINE_float("epsilon_start", 0.1, "Starting exploration parameter.")
flags.DEFINE_float("epsilon_end", 0.1, "Final exploration parameter.")
flags.DEFINE_bool("use_checkpoints", False, "Save/load neural network weights.")
flags.DEFINE_string("checkpoint_dir", "/tmp/dqn_test",
                    "Directory to save/load the agent.")
flags.DEFINE_string(
    "logdir", None,
    "Logging dir to use for TF summary files. If None, the metrics will only "
    "be logged to stderr.")


def main(unused_argv):
  game = factory.create_game_with_setting(FLAGS.game_name, FLAGS.env_setting)
  uniform_policy = policy.UniformRandomPolicy(game)
  mfg_dist = distribution.DistributionPolicy(game, uniform_policy)

  envs = [
      rl_environment.Environment(
          game, mfg_distribution=mfg_dist, mfg_population=p)
      for p in range(game.num_players())
  ]
  info_state_size = envs[0].observation_spec()["info_state"][0]
  num_actions = envs[0].action_spec()["num_actions"]

  hidden_layers_sizes = [int(l) for l in FLAGS.hidden_layers_sizes]
  kwargs = {
      "replay_buffer_capacity": FLAGS.replay_buffer_capacity,
      "min_buffer_size_to_learn": FLAGS.min_buffer_size_to_learn,
      "batch_size": FLAGS.batch_size,
      "learn_every": FLAGS.learn_every,
      "learning_rate": FLAGS.rl_learning_rate,
      "optimizer_str": FLAGS.optimizer_str,
      "loss_str": FLAGS.loss_str,
      "update_target_network_every": FLAGS.update_target_network_every,
      "discount_factor": FLAGS.discount_factor,
      "epsilon_decay_duration": FLAGS.epsilon_decay_duration,
      "epsilon_start": FLAGS.epsilon_start,
      "epsilon_end": FLAGS.epsilon_end,
  }

  # pylint: disable=g-complex-comprehension
  agents = [
      dqn.DQN(idx, info_state_size, num_actions, hidden_layers_sizes, **kwargs)
      for idx in range(game.num_players())
  ]
  joint_avg_policy = rl_agent_policy.JointRLAgentPolicy(
      game, {idx: agent for idx, agent in enumerate(agents)},
      envs[0].use_observation)
  if FLAGS.use_checkpoints:
    for agent in agents:
      if agent.has_checkpoint(FLAGS.checkpoint_dir):
        agent.restore(FLAGS.checkpoint_dir)

  # Metrics writer will also log the metrics to stderr.
  just_logging = FLAGS.logdir is None or jax.host_id() > 0
  writer = metrics.create_default_writer(
      logdir=FLAGS.logdir, just_logging=just_logging)

  # Save the parameters.
  writer.write_hparams(kwargs)

  for ep in range(1, FLAGS.num_train_episodes + 1):
    if ep % FLAGS.eval_every == 0:
      writer.write_scalars(ep, {
          f"agent{i}/loss": float(agent.loss) for i, agent in enumerate(agents)
      })

      initial_states = game.new_initial_states()

      # Exact best response to uniform.
      nash_conv_obj = nash_conv.NashConv(game, uniform_policy)
      writer.write_scalars(
          ep, {
              f"exact_br/{state}": value
              for state, value in zip(initial_states, nash_conv_obj.br_values())
          })

      # DQN best response to uniform.
      pi_value = policy_value.PolicyValue(game, mfg_dist, joint_avg_policy)
      writer.write_scalars(ep, {
          f"dqn_br/{state}": pi_value.eval_state(state)
          for state in initial_states
      })

      if FLAGS.use_checkpoints:
        for agent in agents:
          agent.save(FLAGS.checkpoint_dir)

    for p in range(game.num_players()):
      time_step = envs[p].reset()
      while not time_step.last():
        agent_output = agents[p].step(time_step)
        action_list = [agent_output.action]
        time_step = envs[p].step(action_list)

      # Episode is over, step all agents with final info state.
      agents[p].step(time_step)

  # Make sure all values were written.
  writer.flush()


if __name__ == "__main__":
  app.run(main)



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Meta-regret matching with self-play agents."""
from typing import List

from absl import flags
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
import optax

from open_spiel.python.examples.meta_cfr.matrix_games import utils

FLAGS = flags.FLAGS


def opponent_best_response_strategy(utility):
  opponent_action = jnp.argmin(utility, axis=-1)
  opponent_strategy = jax.nn.one_hot(opponent_action, FLAGS.num_actions)
  return opponent_strategy


def _mlp_forwards(mlp_hidden_sizes: List[int]) -> hk.Transformed:
  """Returns a haiku transformation of the MLP model to be used in optimizer.

  Args:
    mlp_hidden_sizes: List containing size of linear layers.

  Returns:
    Haiku transformation of the RNN network.
  """
  def forward_fn(inputs):
    mlp = hk.nets.MLP(mlp_hidden_sizes, activation=jax.nn.relu, name="mlp")
    return mlp(inputs)
  return hk.transform(forward_fn)


class OptimizerModel:
  """Optimizer model."""

  def __init__(self, learning_rate):
    self.learning_rate = learning_rate

    self.model = _mlp_forwards([64, 16, FLAGS.num_actions])

    self._net_init = self.model.init
    self.net_apply = self.model.apply

    self.opt_update, self.net_params, self.opt_state = None, None, None

  def lr_scheduler(self, init_value):
    schedule_fn = optax.polynomial_schedule(
        init_value=init_value, end_value=0.05, power=1., transition_steps=50)
    return schedule_fn

  def get_optimizer_model(self):
    schedule_fn = self.lr_scheduler(self.learning_rate)
    opt_init, self.opt_update = optax.chain(
        optax.scale_by_adam(), optax.scale_by_schedule(schedule_fn),
        optax.scale(-self.learning_rate))
    rng = jax.random.PRNGKey(10)
    dummy_input = np.random.normal(
        loc=0, scale=10., size=(FLAGS.batch_size, 1, FLAGS.num_actions))
    self.net_params = self._net_init(rng, dummy_input)
    self.opt_state = opt_init(self.net_params)


class MetaSelfplayAgent:
  """Meta player."""

  def __init__(self, repeats, training_epochs, data_loader):
    self.repeats = repeats
    self.training_epochs = training_epochs
    self.net_apply = None
    self.net_params = None
    self.regret_sum = None
    self.step = 0
    self.data_loader = data_loader

  def train(self):
    self.training_optimizer()
    self.regret_sum = jnp.zeros(shape=[FLAGS.batch_size, 1, FLAGS.num_actions])

  def initial_policy(self):
    x = self.net_apply(self.net_params, None, self.regret_sum)
    self.last_policy = jax.nn.softmax(x)
    self.step += 1
    return self.last_policy

  def next_policy(self, last_values):
    value = jnp.matmul(self.last_policy, last_values)
    curren_regret = jnp.transpose(last_values, [0, 2, 1]) - value
    self.regret_sum += curren_regret

    x = self.net_apply(self.net_params, None, self.regret_sum / (self.step + 1))
    self.last_policy = jax.nn.softmax(x)
    self.step += 1
    return self.last_policy

  def training_optimizer(self):
    """Training optimizer."""

    optimizer = OptimizerModel(0.01)
    optimizer.get_optimizer_model()

    for _ in range(FLAGS.num_batches):
      batch_payoff = next(self.data_loader)
      # for _ in range(self.repeats):
      grads = jax.grad(
          utils.meta_loss,
          has_aux=False)(optimizer.net_params, optimizer.net_apply,
                         batch_payoff, self.training_epochs)

      updates, optimizer.opt_state = optimizer.opt_update(
          grads, optimizer.opt_state)
      optimizer.net_params = optax.apply_updates(optimizer.net_params, updates)

    self.net_apply = optimizer.net_apply
    self.net_params = optimizer.net_params
